#include "Born.h"
#include <iostream>
namespace SEP {

BornEngine::BornEngine(Space *_mspace, Space *_uspace){
  mspace = (inCoreFloat *)_mspace;
  uspace = (inCoreFloat *)_uspace;
  setDomain(_mspace);
  setRange(_uspace);
  std::vector<llong> out_dims = uspace->get_ns();
  nt = out_dims[0];
  nz = out_dims[1];
  nx = out_dims[2];
  mblock = nz*nx;
  Dtemp = new float[2*mblock];
}

void BornEngine::init(){
  
  // Set up temporary Vectors
  MK      = new Vector(mspace);
  VTX     = new Vector(mspace);
  MRZ     = new Vector(mspace);
  MRX     = new Vector(mspace);
  MR      = new Vector(mspace);
  temp    = new Vector(mspace);
  tempX   = new Vector(mspace);
  tempZ   = new Vector(mspace);  
  tempuf  = new Vector(mspace);
  tempub  = new Vector(mspace); 
  temp1   = new Vector(mspace);
  temp2   = new Vector(mspace);
   
  // Set up Operators
  DX = new GradX(mspace,mspace, nb);
  DZ = new GradZ(mspace,mspace, nb);
  DXT = new GradXT(mspace,mspace, nb);
  DZT = new GradZT(mspace,mspace, nb);
  interpX = new IntX(mspace,mspace);
  interpZ = new IntZ(mspace,mspace);

}


void BornEngine::SetPar(llong _nb, float _dt, float _dx, llong _sample){
  sample = _sample;
  nb = _nb;
  dt = _dt;
  dx = _dx;
  dt2= 1./(dt*dt);
  dx2= 1./(dx*dx);
  dtdx2=(dt/dx)*(dt/dx);
  init();
}

void BornEngine::SetBG(Vector *_BG){
  BG = _BG;
}

void BornEngine::SetModel(Vector *_MK, Vector *_MR){
  MK->linearCombo(0.,*_MK,1.);
  MR->linearCombo(0.,*_MR,1.);
  float *__restrict__ mkval = ((dataInCoreFloat*)MK->getDataContainer())->getFloatPtr();
  float *__restrict__ cval = ((dataInCoreFloat*)VTX->getDataContainer())->getFloatPtr();

  for(long long i=0; i<nx*nz;i++){
    cval[i] = dtdx2/mkval[i];
  }     // vtx = (dt/dx)^2 v^2 rho
  interpX->Forward(false,MR,MRX);
  interpZ->Forward(false,MR,MRZ);
}

bool BornEngine::FMarch(Vector *pre, Vector *cur, Vector *next){
  float *__restrict__ v1 = ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
  float *__restrict__ v2 = ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr(); 
  float *__restrict__ v3 = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
  float *__restrict__ cval = ((dataInCoreFloat*)VTX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)MRX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)MRZ->getDataContainer())->getFloatPtr();
  float *__restrict__ t = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ tx = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tz = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();

  
  DX->Forward(false,cur,temp);
  
  // ISPC substitution
  ScaleISPC(rinvx_val,t,mblock);
  /*
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvx_val[i];
  }
  */
  DXT->Forward(false,temp,tempX);

  DZ->Forward(false,cur,temp);
  // ISPC substitution
  ScaleISPC(rinvz_val,t,mblock);
  /*
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvz_val[i];
  }
  */
  DZT->Forward(false,temp,tempZ);

  // ISPC substitution  
  LeapFrogISPC(cval,v1,tx,tz,v2,v3,mblock);
  /*
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*(tx[i]+tz[i])-v1[i];
  }
  */
  
  /*
  Constant density case
  Lap->Forward(false,cur,temp);
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*t[i]-v1[i];
  } 
  */ 

  return 0;
}

bool BornEngine::BMarch(Vector *pre, Vector *cur, Vector *next){
  float *__restrict__ v1 = ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
  float *__restrict__ v2 = ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr(); 
  float *__restrict__ v3 = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
  float *__restrict__ cval = ((dataInCoreFloat*)VTX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)MRX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)MRZ->getDataContainer())->getFloatPtr();
  float *__restrict__ t = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ tx = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tz = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();


  DXT->Adjoint(false,temp,cur);
  // ISPC substitution
  ScaleISPC(rinvx_val,t,mblock);
  /*  
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvx_val[i];
  } 
  */ 
  DX->Adjoint(false,tempX,temp);

  DZT->Adjoint(false,temp,cur);
  // ISPC substitution
  ScaleISPC(rinvz_val,t,mblock);
  /*  
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvz_val[i];
  } 
  */ 
  DZ->Adjoint(false,tempZ,temp);
  
  // ISPC substitution  
  LeapFrogISPC(cval,v1,tx,tz,v2,v3,mblock);
  /*
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*(tx[i]+tz[i])-v1[i];
  }
  */  
 
  /*
  Lap->Adjoint(false,temp,cur);
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*t[i]-v1[i];
  } 
  */ 
  return 0;
}
    
void BornEngine::TimeCoreForward(bool add, Vector *model, Vector *data){
  if(!add) data->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ mkval= ((dataInCoreFloat*)MK->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();

  
  for(llong i = 0; i < nz*nx; i++) Dtemp[i] = mkval[i]*m[i]*dt2/(sample*sample);
  for(llong it=1; it < nt-1; it++){
	llong shift = it*mblock;
	// ISPC substitution
	TimeCoreISPC(&uval[shift-mblock],&uval[shift],&uval[shift+mblock],Dtemp,&d[shift],mblock);
	/*
	for(long long i = 0,j=shift;i < mblock; i++,j++) {
	  d[j] += Dtemp[i]*(uval[j-mblock]-2.*uval[j]+uval[j+mblock]);    //      u field
	}
	*/
  } 
}

void BornEngine::TimeCoreAdjoint(bool add, Vector *model, Vector *data){
  if(!add) model->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ mkval= ((dataInCoreFloat*)MK->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();


  for(llong i = 0; i < mblock; i++) Dtemp[i]=0.;
  for(llong it=1; it < nt-1; it++){
	llong shift = it*mblock;
	
	// ISPC substitution
	TimeCoreISPC(&uval[shift-mblock],&uval[shift],&uval[shift+mblock],&d[shift],Dtemp,mblock);
	/*
	for(long long i = 0; i < mblock; i++) {
	  j = i+shift;
	  Dtemp[i] += d[j]*(uval[j-mblock]-2.*uval[j]+uval[j+mblock]);    //      u field
	}
	*/
  } 
  for(llong i = 0; i < mblock; i++) m[i] += dt2*Dtemp[i]*mkval[i]/(sample*sample);
  
}


void BornEngine::SpaceFCoreForward(bool add, Vector *model, Vector *data){
  if(!add) data->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ mrval = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ rinv_val = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)MRX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)MRZ->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();   


  for(llong i=0; i < mblock; i++){
    temp1val[i] = m[i]*mrval[i];
  }

  interpX->Forward(false,temp1,tempX);  // get drho/rho^2 term
  interpZ->Forward(false,temp1,tempZ);  
  
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;

    // Memory OP substitution
    memcpy(tempufval,&uval[shift],sizeof(float)*mblock);
    /*
    for(long long i = 0; i < mblock; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }
    */
    DX->Forward(false,tempuf,temp1);     // first derivative
    DZ->Forward(false,tempuf,temp2);  

    // ISPC substitution
    ScaleISPC(tempxval,temp1val,nz*nx);
    ScaleISPC(tempzval,temp2val,nz*nx);    
   
    /*
    for(long long i = 0; i < nz*nx; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];    //      multiply by drho/rho^2
      temp2val[i] = temp2val[i]*tempzval[i];    //    
    } 
    */
    DXT->Forward(false,temp1,tempuf);           // tilde derivative
    DZT->Forward(false,temp2,tempub);
    for(long long i = 0; i < mblock; i++) {
      d[i + shift] -= (tempufval[i]+tempubval[i])*dx2;    //      u field
    }
  }    
}

void BornEngine::SpaceFCoreAdjoint(bool add, Vector *model, Vector *data){
  if(!add) model->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ mrval = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ rinv_val = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)MRX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)MRZ->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();   
  
   
  for(llong i=0; i < mblock; i++){
    Dtemp[i] = 0.;
  }
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    
    
    // Memory OP substitution
    memcpy(tempufval,&uval[shift],sizeof(float)*mblock);
    /*
    for(long long i = 0; i < mblock; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }
    */
    DX->Forward(false,tempuf,temp1);
    DZ->Forward(false,tempuf,temp2);  
    
    // Memory OP substitution
    memcpy(tempufval,&d[shift],sizeof(float)*mblock);
    /*
    for(long long i = 0; i < mblock; i++) {
      tempufval[i] = d[i + shift];    //      u field
    }
    */
    DXT->Adjoint(false,tempX,tempuf);
    DZT->Adjoint(false,tempZ,tempuf);

    // ISPC substitution
    ScaleISPC(tempxval,temp1val,nz*nx);
    ScaleISPC(tempzval,temp2val,nz*nx);   
    /*
    for(long long i = 0; i < mblock; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];
      temp2val[i] = temp2val[i]*tempzval[i];
    }
    */
    
    
    interpX->Adjoint(false,tempX,temp1);
    interpZ->Adjoint(false,tempZ,temp2);
    for(llong i=0; i < mblock; i++){
      Dtemp[i] += (tempxval[i]+tempzval[i]);
    } 
  } 
  for(llong i=0; i < mblock; i++){
    m[i] -= Dtemp[i]*mrval[i]*dx2;
  }
}

void BornEngine::SpaceACoreForward(bool add, Vector *model, Vector *data){
  if(!add) data->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ mrval = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ rinv_val = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)MRX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)MRZ->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();   


  for(llong i=0; i < mblock; i++){
    temp1val[i] = m[i]*mrval[i];
  }

  interpX->Forward(false,temp1,tempX);  // get drho/rho^2 term
  interpZ->Forward(false,temp1,tempZ);  
  
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    
    // Memory OP substitution
    memcpy(tempufval,&uval[shift],mblock*sizeof(float));
    /*
    for(long long i = 0; i < mblock; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }
    */
    DXT->Adjoint(false,temp1,tempuf);     // first derivative
    DZT->Adjoint(false,temp2,tempuf);  
    // ISPC substitution
    ScaleISPC(tempxval,temp1val,nz*nx);
    ScaleISPC(tempzval,temp2val,nz*nx);     
    /*
    for(long long i = 0; i < nz*nx; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];    //      multiply by drho/rho^2
      temp2val[i] = temp2val[i]*tempzval[i];    //    
    }
    */
    DX->Adjoint(false,tempuf,temp1);           // tilde derivative
    DZ->Adjoint(false,tempub,temp2);
    for(long long i = 0; i < mblock; i++) {
      d[i + shift] -= (tempufval[i]+tempubval[i])*dx2;    //      u field
    }
  }    
}
void BornEngine::SpaceACoreAdjoint(bool add, Vector *model, Vector *data){
  if(!add) model->zeroElement();
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ mrval = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ uval = ((dataInCoreFloat*)BG->getDataContainer())->getFloatPtr();
  float *__restrict__ rinv_val = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)MRX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)MRZ->getDataContainer())->getFloatPtr();
  float *__restrict__ tempufval = ((dataInCoreFloat*)tempuf->getDataContainer())->getFloatPtr();
  float *__restrict__ tempubval = ((dataInCoreFloat*)tempub->getDataContainer())->getFloatPtr();
  float *__restrict__ temp1val = ((dataInCoreFloat*)temp1->getDataContainer())->getFloatPtr();
  float *__restrict__ temp2val = ((dataInCoreFloat*)temp2->getDataContainer())->getFloatPtr();  
  float *__restrict__ tempxval = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tempzval = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();   
   
  for(llong i=0; i < mblock; i++){
    Dtemp[i] = 0.;
  }
  for(llong it=1; it < nt-1; it++){
    llong shift = it*mblock;
    for(long long i = 0; i < mblock; i++) {
      tempufval[i] = uval[i + shift];    //      u field
    }
    DXT->Adjoint(false,temp1,tempuf);
    DZT->Adjoint(false,temp2,tempuf);  
    for(long long i = 0; i < mblock; i++) {
      tempufval[i] = d[i + shift];    //      u field
    }
    DX->Forward(false,tempuf,tempX);
    DZ->Forward(false,tempuf,tempZ);
    for(long long i = 0; i < mblock; i++) {
      temp1val[i] = temp1val[i]*tempxval[i];
      temp2val[i] = temp2val[i]*tempzval[i];
    }
    interpX->Adjoint(false,tempX,temp1);
    interpZ->Adjoint(false,tempZ,temp2);
    for(llong i=0; i < mblock; i++){
      Dtemp[i] += (tempxval[i]+tempzval[i]);
    } 
  } 
  for(llong i=0; i < mblock; i++){
    m[i] -= Dtemp[i]*mrval[i]*dx2;
  }
}



bool BornEngine::Forward(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(data,mblock,nt);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  return 0;
}

bool BornEngine::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nx,nz);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  return 0;
}

bool BornEngine::StarForward(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(data,mblock,nt);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  return 0;
}

bool BornEngine::StarAdjoint(bool add, Vector *model, Vector *data){
  if(!add) Zero.Zero(model,nx,nz);
  float *__restrict__ m = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ d = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();

  return 0;
}


/*
 The following is Born Operator for time derivative
*/
BornT::BornT(Space *_mspace, Space *_uspace){
  setDomain(_mspace);
  setRange(_uspace);
}
void BornT::SetEngine(BornEngine *_BornMod){
  BornMod = _BornMod;
}
bool BornT::Forward(bool add, Vector *model, Vector *data){
  BornMod->TimeCoreForward(add,model,data);
}
bool BornT::Adjoint(bool add, Vector *model, Vector *data){
  BornMod->TimeCoreAdjoint(add,model,data);
}
    
/*
 The following is Born Operator for m_rho derivative
*/
BornMR::BornMR(Space *_mspace, Space *_uspace){
  setDomain(_mspace);
  setRange(_uspace);
}
void BornMR::SetEngine(BornEngine *_BornMod){
  BornMod = _BornMod;
}
bool BornMR::Forward(bool add, Vector *model, Vector *data){
  BornMod->SpaceACoreForward(add,model,data);
}
bool BornMR::Adjoint(bool add, Vector *model, Vector *data){
  BornMod->SpaceACoreAdjoint(add,model,data);
}
    
}