#include <cmath>
#include <iostream>
#include <algorithm>

#include "Engine.h"

namespace SEP {

void Engine::TestEcho(){
  std::cout << "Hello World!" << std::endl;
  LVec[0]->DotTest(true);
}

void Engine::SetSpace(Space *_mspace, Space *_dspace, Space *_uspace){
  mspace = (inCoreFloat *)_mspace;
  dspace = (inCoreFloat *)_dspace;
  uspace = (inCoreFloat *)_uspace;
  MK = new Vector(mspace);
  MR = new Vector(mspace);
}

void Engine::SetShot(Vector *_shot, std::vector<int> _sxv, std::vector<int> _szv){
  shot = new Vector(_shot);
  sxv  = _sxv;
  szv  = _szv;
}

void Engine::SetPropPar(llong _nb, llong _bsize, llong _boundary, float _dt, float _dx, float _dz, int _sample, int _water, float *_weight){
  nb      = _nb;
  bsize   = _bsize;
  boundary= _boundary;
  dt      = _dt;
  dx      = _dx;
  dz      = _dz;
  sample  = _sample;
  water   = _water;
  weight  = _weight;
  init();

}

void Engine::init(){

  std::vector<long long> in_dims = mspace->get_ns();
  std::vector<long long> out_dims = dspace->get_ns();
  
  // Dimensions of parameters m[nz][nx] and u[nt][nz][nx]
  nz = in_dims[0];
  nx = in_dims[1];
  ns = sxv.size();
  nt = out_dims[0];
  if(weight == NULL) {
    weight = new float[nx*nz];
    for(int i=0;i<nx*nz;i++){
      weight[i] = 1.;
    }
  }
  pi = 3.14159265358979323846264338327950288419716939937510582;
  Vector *temp = new Vector(mspace);
  dataTrue = new float[nx*ns*nt];
  dataRes  = new float[nx*ns*nt];
  // Create Propagation modules and Vectors
  for(int i=0; i < nb; i++){
    // Born Operator
    BornVec.push_back(new BornEngine(mspace,uspace));
    BornVec[i]->SetPar(boundary, dt, dx, sample);
    // propagators
    LVec.push_back(new Prop(uspace,uspace));
    LVec[i]->SetPropPar(BornVec[i],boundary,dt,dx,temp,sample);
    //Source Vector
    SourceVec.push_back(new Source(nt,nz,nx,sample));
    ShotVec.push_back(new Vector(shot));
    //SensorPad Vector
    PadVec.push_back(new SensorPad(dspace,uspace,szv[0]));
    // Vectors for propagation
    FU0Vec.push_back(new Vector(uspace));
    FU1Vec.push_back(new Vector(uspace));
    AU0Vec.push_back(new Vector(uspace));
    AU1Vec.push_back(new Vector(uspace)); 
    SVec.push_back(new Vector(uspace));   
    dMKVec.push_back(new Vector(mspace));
    dMRVec.push_back(new Vector(mspace));
    dmVVec.push_back(new Vector(mspace));
    dmrVec.push_back(new Vector(mspace));
    C1Vec.push_back(new Vector(mspace));
    C2Vec.push_back(new Vector(mspace));
    //C3Vec.push_back(new Vector(mspace));
    //C4Vec.push_back(new Vector(mspace));
  }
  
  // Create Data
  for(int i=0; i < ns; i++){
    DataVec.push_back(new Vector(dspace));
    Data1Vec.push_back(new Vector(dspace));
  }
}

void Engine::SetTrueModel(Vector *_MK, Vector *_MR){
  MK->linearCombo(0.,*_MK,1.);
  MR->linearCombo(0.,*_MR,1.); 
  for(int i = 0; i < nb; i++){
    LVec[i]->SetModel(MK,MR);
  }
  int ib, id, is;
  #pragma omp parallel for default(shared) private(ib,id,is)
  for(ib=0; ib < nb; ib++){
    for(id=0;id<bsize;id++){
      is = ib*bsize+id;
      // source injection
      SVec[ib]->zeroElement();
      SourceVec[ib]->Injection(ShotVec[ib],SVec[ib],szv[is],sxv[is]);
      // wave propagation
      LVec[ib]->Forward(false,SVec[ib],FU0Vec[ib]);
      //data collection
      PadVec[ib]->SetShot(sxv[is]);
      PadVec[ib]->Adjoint(false,DataVec[is],FU0Vec[ib]);
    }
  }  
  for(llong i=0; i < ns; i++){
    float *__restrict__ d_val = ((dataInCoreFloat*)DataVec[i]->getDataContainer())->getFloatPtr(); 
    for(llong j=0; j < nt*nx; j++){
      dataTrue[i*nt*nx+j]=d_val[j];
    }
  }
}


void Engine::SetModel(Vector *_MK, Vector *_MR){
  MK->linearCombo(0.,*_MK,1.);
  MR->linearCombo(0.,*_MR,1.);  
  for(int i = 0; i < nb; i++){
    LVec[i]->SetModel(MK,MR);
  }
  
}
double Engine::OBJF(){
  double result = 0.;
  for(llong i=0; i < ns*nt*nx; i++){
      result +=dataRes[i]*dataRes[i];
  }
}
void Engine::CreateData(){
  int ib, id, is;
  
  #pragma omp parallel for default(shared) private(ib,id,is)
  for(ib=0; ib < nb; ib++){
    for(id=0;id<bsize;id++){
      is = ib*bsize+id;
      // source injection
      SVec[ib]->zeroElement();
      SourceVec[ib]->Injection(ShotVec[ib],SVec[ib],szv[is],sxv[is]);
      // wave propagation
      LVec[ib]->Forward(false,SVec[ib],FU0Vec[ib]);
      //data collection
      PadVec[ib]->SetShot(sxv[is]);
      PadVec[ib]->Adjoint(false,DataVec[is],FU0Vec[ib]);
    }
  }
  for(llong i=0; i < ns; i++){
    float *__restrict__ d_val = ((dataInCoreFloat*)DataVec[i]->getDataContainer())->getFloatPtr(); 
    for(llong j=0; j < nt*nx; j++){
      dataRes[i*nt*nx+j]=d_val[j]-dataTrue[i*nt*nx+j];
      d_val[j] = dataRes[i*nt*nx+j];//-dataTrue[i*nt*nx+j];
    }
  }  
}

void Engine::GetData(float *data){
  for(llong i=0; i < ns; i++){
    float *__restrict__ d_val = ((dataInCoreFloat*)DataVec[i]->getDataContainer())->getFloatPtr(); 
    for(llong j=0; j < nt*nx; j++){
      data[i*nt*nx+j]=d_val[j];
    }
  }
}
void Engine::GetTrueData(float *data){
  for(llong i=0; i < ns*nt*nx; i++){
      data[i]=dataTrue[i];
  }
}
void Engine::PutData(float *data){
  for(llong i=0; i < ns; i++){
    DataVec[i]->zeroElement();
    float *__restrict__ d_val = ((dataInCoreFloat*)DataVec[i]->getDataContainer())->getFloatPtr(); 
    for(llong j=0; j < nt*nx; j++){
      d_val[j]=data[i*nt*nx+j];
    }
  }
}
void Engine::GetFWave(Vector *wave, int i){
  wave->linearCombo(0.,*FU0Vec[i],1.);

}
void Engine::GetAdjWave(Vector *wave, int i){
  wave->linearCombo(0.,*AU0Vec[i],1.);

}

void Engine::GradientFunc(Vector *GradMK, Vector *GradMR){

  GradMK->zeroElement();
  GradMR->zeroElement();
  int ib, id, is;
  float *dval;
  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib=0; ib < nb; ib++){
    dMKVec[ib]->zeroElement();
    dMRVec[ib]->zeroElement();
  }
  
  #pragma omp parallel for default (shared) private(ib,id,is,dval)
  for(ib=0; ib < nb; ib++){
    for(id=0; id < bsize; id++){
      is=ib*bsize+id;
      //Source Injection
      SVec[ib]->zeroElement();
      SourceVec[ib]->Injection(ShotVec[ib],SVec[ib],szv[is],sxv[is]);
      // Forward Simulation
      LVec[ib]->Forward(false,SVec[ib],FU0Vec[ib]);
      
      // Injection data residual
      PadVec[ib]->SetShot(sxv[is]);
      PadVec[ib]->Adjoint(false,DataVec[is],FU0Vec[ib]);
      dval = ((dataInCoreFloat*)DataVec[is]->getDataContainer())->getFloatPtr(); 
      for(llong j=0; j < nt*nx; j++){
        dataRes[is*nt*nx+j]=dval[j]-dataTrue[is*nt*nx+j];
        dval[j] = dataRes[is*nt*nx+j];//-dataTrue[i*nt*nx+j];
      }      
      PadVec[ib]->Forward(false,DataVec[is], SVec[ib]);
      
      // Backward Simulation
      LVec[ib]->Adjoint(false,AU0Vec[ib],SVec[ib]);
      
      // Imaging condition
      BornVec[ib]->SetBG(FU0Vec[ib]);
      BornVec[ib]->TimeCoreAdjoint(true,dMKVec[ib],AU0Vec[ib]);
      BornVec[ib]->SpaceFCoreAdjoint(true,dMRVec[ib],AU0Vec[ib]);
    }
  }

    float *__restrict__ GMK_val = ((dataInCoreFloat*)GradMK->getDataContainer())->getFloatPtr();
    float *__restrict__ GMR_val = ((dataInCoreFloat*)GradMR->getDataContainer())->getFloatPtr();
    float *__restrict__ MK_val = ((dataInCoreFloat*)MK->getDataContainer())->getFloatPtr();
    float *__restrict__ MR_val = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
    for(int i = 0; i < nb; i++){
      float *__restrict__ dMK_val = ((dataInCoreFloat*)dMKVec[i]->getDataContainer())->getFloatPtr();
      float *__restrict__ dMR_val = ((dataInCoreFloat*)dMRVec[i]->getDataContainer())->getFloatPtr();

      for(int j = water; j < nz*nx; j++){
        GMK_val[j] += dMK_val[j]*weight[j];
        GMR_val[j] += dMR_val[j]*weight[j];
      }
    }
  
} // End of GradientFunc 

void Engine::HessianFunc(Vector *dMK, Vector *dMR, Vector *HMK, Vector *HMR){
  // we do not clean HV, Hr in this function
  int ib,id,is;

  float *__restrict__ dMK_val = ((dataInCoreFloat*)dMK->getDataContainer())->getFloatPtr();
  float *__restrict__ dMR_val = ((dataInCoreFloat*)dMR->getDataContainer())->getFloatPtr();
  
  for(ib = 0; ib < nb; ib++){
    float *__restrict__ dmK_val = ((dataInCoreFloat*)dMKVec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    float *__restrict__ dmrho_val = ((dataInCoreFloat*)dMRVec[ib]->getDataContainer())->getFloatPtr();
    for(llong i=0; i < water; i++){
      dmrho_val[i] = 0.;
      dmK_val[i] = 0.; 
    }   
    for(llong i=water; i < nx*nz; i++){
      dmrho_val[i] = dMR_val[i]*weight[i];
      dmK_val[i] = dMK_val[i]*weight[i]; 
    }   
    //dMKVec[ib]->zeroElement();
    //dMRVec[ib]->zeroElement();
    C1Vec[ib]->zeroElement();
    C2Vec[ib]->zeroElement();
    //C3Vec[ib]->zeroElement();
    //C4Vec[ib]->zeroElement();
  }

  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib = 0; ib < nb; ib++){
    //std::cout << "working at node :" << ib <<std::endl;
    for(id = 0; id < bsize; id++){
      is = ib*bsize+id;
      PadVec[ib]->SetShot(sxv[is]);
      // Source injection
      //SVec[ib]->zeroElement();
      //SourceVec[ib]->Injection(ShotVec[ib], SVec[ib], szv[is], sxv[is]);

      // Forward simulation
      //LVec[ib]->Forward(false,SVec[ib],FU0Vec[ib]);        // U for background        
      // Back simulation
      //PadVec[ib]->Forward(false,DataVec[is],SVec[ib]);   
      //LVec[ib]->Adjoint(false,AU0Vec[ib],SVec[ib]);        // VVec = lambda
 
      // Born approximation of Forward  
      BornVec[ib]->SetBG(FU0Vec[ib]);  
      BornVec[ib]->TimeCoreForward(false,dMKVec[ib],SVec[ib]);    // time derivative term 
      BornVec[ib]->SpaceFCoreForward(true,dMRVec[ib],SVec[ib]);  // space derivative term
      LVec[ib]->Forward(false,SVec[ib],FU1Vec[ib]);
      
      // Born approxi + 2nd order scatter of Adjoint
      BornVec[ib]->SetBG(AU0Vec[ib]);
      BornVec[ib]->TimeCoreForward(false,dMKVec[ib],SVec[ib]);
      BornVec[ib]->SpaceACoreForward(true,dMRVec[ib],SVec[ib]);  // space derivative term
      PadVec[ib]->Adjoint(false,Data1Vec[is],FU1Vec[ib]);
      PadVec[ib]->Forward(true,Data1Vec[is],SVec[ib]);
      
      // Backward propagation
      LVec[ib]->Adjoint(false,AU1Vec[ib],SVec[ib]);
       
      // Imaging condition
      BornVec[ib]->SetBG(FU0Vec[ib]);  
      BornVec[ib]->TimeCoreAdjoint(true,C1Vec[ib],AU1Vec[ib]);
      BornVec[ib]->SpaceFCoreAdjoint(true,C2Vec[ib],AU1Vec[ib]);  // space derivative term
      BornVec[ib]->SetBG(FU1Vec[ib]);  
      BornVec[ib]->TimeCoreAdjoint(true,C1Vec[ib],AU0Vec[ib]);
      BornVec[ib]->SpaceFCoreAdjoint(true,C2Vec[ib],AU0Vec[ib]);  // space derivative term

    }
  }

  float *__restrict__ HMK_val = ((dataInCoreFloat*)HMK->getDataContainer())->getFloatPtr();   // coef for time Born term
  float *__restrict__ HMR_val = ((dataInCoreFloat*)HMR->getDataContainer())->getFloatPtr();
  float *__restrict__ MK_val = ((dataInCoreFloat*)MK->getDataContainer())->getFloatPtr();
  float *__restrict__ MR_val = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  for(ib = 0; ib < nb; ib++){
    float *__restrict__ C1_val = ((dataInCoreFloat*)C1Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    float *__restrict__ C2_val = ((dataInCoreFloat*)C2Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    //float *__restrict__ C3_val = ((dataInCoreFloat*)C3Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    //float *__restrict__ C4_val = ((dataInCoreFloat*)C4Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    for(llong i=water; i < nx*nz; i++){
      HMK_val[i] += C1_val[i]*weight[i];
      HMR_val[i] += C2_val[i]*weight[i];
    }   
  }
  for(llong i=0; i < nx*nz; i++){
      HMK_val[i] = HMK_val[i];//+dV_val[i]*0.01;
      HMR_val[i] = HMR_val[i];//+dr_val[i]*0.01;
    }   
    
}

void Engine::GaussHessianFunc(Vector *dMK, Vector *dMR, Vector *HMK, Vector *HMR){
  // we do not clean HV, Hr in this function
  int ib,id,is;

  float *__restrict__ dMK_val = ((dataInCoreFloat*)dMK->getDataContainer())->getFloatPtr();
  float *__restrict__ dMR_val = ((dataInCoreFloat*)dMR->getDataContainer())->getFloatPtr();
  
  for(ib = 0; ib < nb; ib++){
    float *__restrict__ dmK_val = ((dataInCoreFloat*)dMKVec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    float *__restrict__ dmrho_val = ((dataInCoreFloat*)dMRVec[ib]->getDataContainer())->getFloatPtr();
    for(llong i=0; i < water; i++){
      dmrho_val[i] = 0.;
      dmK_val[i] = 0.; 
    }   
    for(llong i=water; i < nx*nz; i++){
      dmrho_val[i] = dMR_val[i]*weight[i];
      dmK_val[i] = dMK_val[i]*weight[i]; 
    }   
    C1Vec[ib]->zeroElement();
    C2Vec[ib]->zeroElement();
  }

  #pragma omp parallel for default (shared) private(ib,id,is)
  for(ib = 0; ib < nb; ib++){
    //std::cout << "working at node :" << ib <<std::endl;
    for(id = 0; id < bsize; id++){
      is = ib*bsize+id;
      PadVec[ib]->SetShot(sxv[is]);
      // Source injection
      //SVec[ib]->zeroElement();
      //SourceVec[ib]->Injection(ShotVec[ib], SVec[ib], szv[is], sxv[is]);

      // Forward simulation
      //LVec[ib]->Forward(false,SVec[ib],FU0Vec[ib]);        // U for background        
      // Back simulation
      //PadVec[ib]->Forward(false,DataVec[is],SVec[ib]);   
      //LVec[ib]->Adjoint(false,AU0Vec[ib],SVec[ib]);        // VVec = lambda
 
      // Born approximation of Forward  
      BornVec[ib]->SetBG(FU0Vec[ib]);  
      BornVec[ib]->TimeCoreForward(false,dMKVec[ib],SVec[ib]);    // time derivative term 
      BornVec[ib]->SpaceFCoreForward(true,dMRVec[ib],SVec[ib]);  // space derivative term
      LVec[ib]->Forward(false,SVec[ib],FU1Vec[ib]);
      
      // Born approxi + 2nd order scatter of Adjoint
      PadVec[ib]->Adjoint(false,Data1Vec[is],FU1Vec[ib]);
      PadVec[ib]->Forward(false,Data1Vec[is],SVec[ib]);
      
      // Backward propagation
      LVec[ib]->Adjoint(false,AU1Vec[ib],SVec[ib]);
       
      // Imaging condition
      BornVec[ib]->SetBG(FU0Vec[ib]);  
      BornVec[ib]->TimeCoreAdjoint(true,C1Vec[ib],AU1Vec[ib]);
      BornVec[ib]->SpaceFCoreAdjoint(true,C2Vec[ib],AU1Vec[ib]);  // space derivative term

    }
  }

  float *__restrict__ HMK_val = ((dataInCoreFloat*)HMK->getDataContainer())->getFloatPtr();   // coef for time Born term
  float *__restrict__ HMR_val = ((dataInCoreFloat*)HMR->getDataContainer())->getFloatPtr();
  float *__restrict__ MK_val = ((dataInCoreFloat*)MK->getDataContainer())->getFloatPtr();
  float *__restrict__ MR_val = ((dataInCoreFloat*)MR->getDataContainer())->getFloatPtr();
  for(ib = 0; ib < nb; ib++){
    float *__restrict__ C1_val = ((dataInCoreFloat*)C1Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    float *__restrict__ C2_val = ((dataInCoreFloat*)C2Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    //float *__restrict__ C3_val = ((dataInCoreFloat*)C3Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    //float *__restrict__ C4_val = ((dataInCoreFloat*)C4Vec[ib]->getDataContainer())->getFloatPtr();   // coef for time Born term
    for(llong i=water; i < nx*nz; i++){
      HMK_val[i] += C1_val[i]*weight[i];
      HMR_val[i] += C2_val[i]*weight[i];
    }   
  }   
}

Hess::Hess(Space *_Vspace, Space *mod, Space *data, Engine *_Run){
  setDomain(mod);
  setRange(data);
  Vspace = (inCoreFloat *)_Vspace;
  mspace = (inCoreFloat *)mod;
  Run = _Run;
  Vtemp1 = new Vector(Vspace);
  Vtemp2 = new Vector(Vspace);
  rtemp1 = new Vector(Vspace);
  rtemp2 = new Vector(Vspace);
  mblock = Vspace->get_n123();
  std::cout << "mblock = " << mblock << std::endl;
}

bool Hess::Forward(bool add, Vector *model, Vector *data){
  if(!add) data->zeroElement();
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();  
  float *__restrict__ V1val = ((dataInCoreFloat*)Vtemp1->getDataContainer())->getFloatPtr();
  float *__restrict__ r1val = ((dataInCoreFloat*)rtemp1->getDataContainer())->getFloatPtr(); 
  float *__restrict__ V2val = ((dataInCoreFloat*)Vtemp2->getDataContainer())->getFloatPtr();
  float *__restrict__ r2val = ((dataInCoreFloat*)rtemp2->getDataContainer())->getFloatPtr();
  for(int i=0; i < mblock; i++){
	V1val[i] = mval[i];
    r1val[i] = mval[i+mblock];
  }
  Vtemp2->zeroElement();
  rtemp2->zeroElement();       
  Run->GaussHessianFunc(Vtemp1,rtemp1,Vtemp2,rtemp2);       
  for(int i=0; i < mblock; i++){
    dval[i] += V2val[i];
    dval[i+mblock] += r2val[i];//+10*mval[i+size]*maskval[i]*maskval[i];
  }       
}

bool Hess::Adjoint(bool add, Vector *model, Vector *data){
  if(!add) model->zeroElement();
  float *__restrict__ mval = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ dval = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();  
  float *__restrict__ V1val = ((dataInCoreFloat*)Vtemp1->getDataContainer())->getFloatPtr();
  float *__restrict__ r1val = ((dataInCoreFloat*)rtemp1->getDataContainer())->getFloatPtr(); 
  float *__restrict__ V2val = ((dataInCoreFloat*)Vtemp2->getDataContainer())->getFloatPtr();
  float *__restrict__ r2val = ((dataInCoreFloat*)rtemp2->getDataContainer())->getFloatPtr();
  for(int i=0; i < mblock; i++){
	V1val[i] = dval[i];
    r1val[i] = dval[i+mblock];
  }
  Vtemp2->zeroElement();
  rtemp2->zeroElement();       
  Run->GaussHessianFunc(Vtemp1,rtemp1,Vtemp2,rtemp2);       
  for(int i=0; i < mblock; i++){
    mval[i] += V2val[i];
    mval[i+mblock] += r2val[i];//+10*mval[i+size]*maskval[i]*maskval[i];
  }       
}



}