#ifndef HESSINV_H
#define HESSINV_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
#include "Engine.h"
#include "Wrapper.h"
#include <omp.h>
#include "utilities.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
typedef long long llong;

/*
  Gradient operator
  
  GradX:  regular     -> X staggered grid
  GradZ:  regular     -> Z staggered grid
  GradXT: X staggered -> regular
  GradZT: Z staggered -> regular

  
  First Created: 2015-11-03, Yinbin Ma
  Last Modified: 2015-11-03
  
*/


namespace SEP{

class HessInv{
  public:
    HessInv(Space *mod, Engine *Run, Wrapper *Wrap);
    ~HessInv(){}
    bool Apply(Vector *Gv, Vector *Gr, Vector *Dv, Vector *Dr, int niter);
    
  private:
    MKL_INT mblock, n, nm, rci_request, itercount, expected_itercount,i;
    MKL_INT length;
    MKL_INT *ipar;
    double *solution, *rhs, *tmp, *dpar;
    float *resid;
    inCoreFloat *mspace;
    Vector *Hr, *HV;
    Engine *Run;
    Wrapper *Wrap;
    void ConvertFtoD(float *input, double *output, int size);
    void ConvertDtoF(double *input, float *output, int size);
    double OBJF(float *resid,int n);

};


HessInv::HessInv(Space *mod, Engine *_Run,Wrapper *_Wrap){
  mspace = (inCoreFloat *)mod;
  Run = _Run;
  Wrap = _Wrap;
  mblock= ((inCoreFloat *)mod)->get_n123();
  nm=mblock;
  n = 2*mblock;
  length=128;
  solution = new double[n];
  rhs = new double[n];
  resid = new float[n];
  tmp = new double[n*4];
  dpar= new double[128];
  ipar = new MKL_INT[128];
  HV = new Vector(mspace);
  Hr = new Vector(mspace);
  std::cout << "mblock = " << mblock <<", nm = " << nm <<std::endl;

}

bool HessInv::Apply(Vector *GV, Vector *Gr, Vector *dV, Vector *dr, int niter){
  float *__restrict__ GVval = ((dataInCoreFloat*)GV->getDataContainer())->getFloatPtr();
  float *__restrict__ Grval = ((dataInCoreFloat*)Gr->getDataContainer())->getFloatPtr();
  float *__restrict__ dVval = ((dataInCoreFloat*)dV->getDataContainer())->getFloatPtr();
  float *__restrict__ drval = ((dataInCoreFloat*)dr->getDataContainer())->getFloatPtr();
  float *__restrict__ HVval = ((dataInCoreFloat*)HV->getDataContainer())->getFloatPtr();
  float *__restrict__ Hrval = ((dataInCoreFloat*)Hr->getDataContainer())->getFloatPtr();
    
  ConvertFtoD(GVval,rhs,nm);
  ConvertFtoD(Grval,&rhs[nm],nm);
  for(int i=0; i < n;i++){
    solution[i] = 0.;
    tmp[i*4+0] = 0.;
    tmp[i*4+1] = 0.;
    tmp[i*4+2] = 0.;
    tmp[i*4+3] = 0.;
  }
  dcg_init(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
  ipar[8] = 1;
  ipar[9] = 0;
  dpar[0] = 1.e-6;
  dcg_check(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
  dV->zeroElement(); dr->zeroElement();
  HV->zeroElement(); Hr->zeroElement();
  double rtol = 1.e-6;
  double t0,t1;
  for(int iter = 0 ; iter < niter; iter++){
    t0 = timer();
    std::cout<<"iter = " << iter ;
    dcg(&n,solution,rhs,&rci_request,ipar,dpar,tmp);
    ConvertDtoF(tmp,dVval,nm);
    ConvertDtoF(&tmp[nm],drval,nm);
    if(rci_request == 0){
      break;
    }
    Wrap->Hessian(dV,dr,HV,Hr);
    ConvertFtoD(HVval,&tmp[n],nm);
    ConvertFtoD(Hrval,&tmp[n+nm],nm);

    ConvertDtoF(&tmp[2*n],resid,n);
    std::cout << ", HV norm = " << HV->innerProd(*HV) ;
    std::cout << ", residue = " << OBJF(resid,n) ;
    HV->zeroElement(); Hr->zeroElement();
    t1=timer();
    std::cout <<", used " << t1-t0 << " seconds." << std::endl;
    
  }
  dcg_get(&n,solution,rhs,&rci_request,ipar,dpar,tmp,&itercount);
  dV->zeroElement(); dr->zeroElement();
  ConvertDtoF(solution,dVval,nm);
  ConvertDtoF(&solution[nm],drval,nm);
  return 0;
}
void HessInv::ConvertFtoD(float *input, double *output, int size){
  for(int i=0; i < size; i++){
    output[i] = (double)input[i];
  }
}
void HessInv::ConvertDtoF(double *input, float *output, int size){
  for(int i=0; i < size; i++){
    output[i] = (float)input[i];
  }
}
double HessInv::OBJF(float *resid,int n){
  double temp = 0.;
  for(int i=0; i < n; i++){
    temp += resid[i]*resid[i];
  }
  return temp;
}






} // end of namespace
#endif