#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "Laplacian.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "CreateModel.h"
#include "HessInv.h"
#include "Wrapper.h"
#include "NCG.h"
#include "utilities.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;



/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "              Test Modulus.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(6);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;             // size of computational grid
  int sx, sz, ns, oxs, dxs;   // source geometry
  int boundary;               // size of zero boudary
  float dx, dz, dt;           // propagation parameter
  int nb, bsize;              // nb is the number of cores, bsize is shot per core
  int sample;                 // data sampling
  int mblock;                 // size of velocity model
  int type;                   // type of inversion, 1 for (V,r=const), 2 for (V=const,I), 10 for joint inversion
  int niter;                  // number of iteration for CG method

   // Get parameters
  getch("nb","d",&nb);               
  getch("type","d",&type);      
  getch("bsize","d",&bsize);        ns = nb*bsize;
  getch("nx","d",&nx);              
  getch("nz","d",&nz);              mblock=nx*nz;         
  getch("nt","d",&nt);              
  getch("boundary","d",&boundary);  
  getch("dx","f",&dx);              dz = dx;
  getch("sz","d",&sz);              
  getch("oxs","d",&oxs);            
  getch("dxs","d",&dxs);            
  getch("dt","f",&dt);              
  getch("ns","d",&ns);             
  getch("sample","d",&sample);      
  getch("niter","d",&niter);  
  // Summary of jobs
  cout << "========================== Beginning of FWI for Marmousi model ==========================" <<endl;  
  cout << "Number of cores = "   << nb << endl;
  cout << "Jobs on each core = " << bsize << endl;
  cout << "Number of shots = "   << ns << endl;
  if(type==1)  cout << "Velocity estimation for constant rho" << endl;
  if(type==2)  cout << "Impedance estimation for constant velocity" << endl;
  if(type==10) cout << "Joint parameter estimation" << endl;  
  cout << "nx = " << nx << endl;
  cout << "nz = " << nz << endl;
  cout << "nt = " << nt << endl;
  cout << "boundary = " << boundary << endl;
  cout << "dx = " << dx <<"[m]"<< endl;  
  cout << "sz = " << sz << endl;
  cout << "oxs = " << oxs << endl;
  cout << "dxs = " << dxs << endl;
  cout << "dt = " << dt << "[ms]"<< endl;
  cout << "sample = " << sample << endl;
  cout << "Number of CG steps = " << niter << endl;
  
  // Set up space 
  inCoreFloat1D *sspace = new inCoreFloat1D(nt);               // space for source
  inCoreFloat2D *mspace = new inCoreFloat2D(nz,nx);            // space for velocity, reflectivity
  inCoreFloat2D *bmspace= new inCoreFloat2D(2*nz,nx);          // space for velocity, reflectivity
  inCoreFloat3D *uspace = new inCoreFloat3D(nt,nz,nx);         // space for wavefield
  inCoreFloat2D *dspace = new inCoreFloat2D(nt,nx);            // space for data
 
  
  // Set up vectors
  Vector *shot     = new Vector(sspace);          // shot vector
  Vector *VT       = new Vector(mspace);          // true velocity
  Vector *RT       = new Vector(mspace);          // true density
  Vector *VM       = new Vector(mspace);          // starting velocity
  Vector *RM       = new Vector(mspace);          // starting density
  Vector *u1       = new Vector(uspace);          // wave field
  Vector *MKT      = new Vector(mspace);          // K_inv for the true model
  Vector *MRT      = new Vector(mspace);          // rho_inv for the true model
  Vector *MKM      = new Vector(mspace);          // K_inv for the starting model
  Vector *MRM      = new Vector(mspace);          // rho_inv for the starting model
 
  // Creating pointers
  float *__restrict__ shot_val  = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VT_val    = ((dataInCoreFloat*)VT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VM_val    = ((dataInCoreFloat*)VM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RT_val    = ((dataInCoreFloat*)RT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RM_val    = ((dataInCoreFloat*)RM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKT_val   = ((dataInCoreFloat*)MKT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKM_val   = ((dataInCoreFloat*)MKM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRT_val   = ((dataInCoreFloat*)MRT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRM_val   = ((dataInCoreFloat*)MRM->getDataContainer())->getFloatPtr();   
  float *__restrict__ u1_val    = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();  
 
  // Read in velocity and density
  cout << "========================== Reading in velocity and density model ==========================" <<endl;  
  sreed("wavelet",shot_val,sizeof(float)*nt); 
  sreed("VT",VT_val,sizeof(float)*nz*nx);
  sreed("VM",VM_val,sizeof(float)*nz*nx);
  sreed("RT",RT_val,sizeof(float)*nz*nx);
  sreed("RM",RM_val,sizeof(float)*nz*nx);

  for(int i=0;i<mblock;i++){
    MKT_val[i] = 1./(RT_val[i]*VT_val[i]*VT_val[i]);
    MKM_val[i] = 1./(RM_val[i]*VM_val[i]*VM_val[i]);  
    MRT_val[i] = 1./RT_val[i];
    MRM_val[i] = 1./RM_val[i];
  }

  int water =0;
  // creating weighting function
  float *weight = new float[mblock];
  for(int iz=0;iz < nz;iz++){
    double coef = exp(-1.*(iz-70)*(iz-70)/100.);
    //coef = 1.;
    if(iz >= 70) coef = 1.;
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      weight[pos] = coef;
    }
  }
  //srite("weight",weight,sizeof(float)*nx*nz);

  //Constructing operators
  cout << "========================== Constructing operator ==========================" <<endl; 
  float *cmodl = new float[2*mblock];
  for(int i=0; i < mblock; i++){
    cmodl[i] =MKM_val[i];
    cmodl[i+mblock] =MRM_val[i];
  }   
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int i=0; i < nb*bsize; i++){
    sxv.push_back(oxs+dxs*i);
    szv.push_back(sz);
  }
  Engine *Run = new Engine();
  Run->SetSpace(mspace, dspace, uspace);
  Run->SetShot(shot,sxv,szv);
  Run->SetPropPar(nb,bsize,boundary,dt,dx,dz,sample,water,weight);
  Run->SetTrueModel(MKT,MRT);
  float *dataTrue= new float[nt*nx*ns]; 
  float *dataRes = new float[nt*nx*ns];    
  Run->GetData(dataRes);
  srite("dataTrue",dataRes,sizeof(float)*nt*nx*ns);
  
  Wrapper *Wrap;
  Wrap = new VRWrapper(mspace,mblock,Run);
  Wrap->SetO(cmodl);                         // set new reference point 
  Wrap->SetType(type);                       // set inversion point

  NCG *ncg = new NCG(mblock, Wrap);
  ncg->SetPar(niter);
  
  
  // Collecting data
  cout << "========================== Collecting data ==========================" <<endl; 
 
  Run->SetModel(MKM,MRM);
  Run->CreateData();
  Run->GetData(dataRes);
  for(int i=0;i<nt*nx*ns;i++){
    dataRes[i] = dataRes[i] + dataTrue[i];
  }
  srite("dataInit",dataRes,sizeof(float)*nt*nx*ns);
  //return 0;
  cout << "========================== Computating gradient ==========================" <<endl; 
  /*
  Run->GradientFunc(dV,dR);
  Run->GetFWave(u1,0);
  for(int i=0;i<mblock;i++){
    dV_val[i] = dV_val[i]/MKM_val[i];
    dR_val[i] = dR_val[i]/MRM_val[i];  
  }

  srite("GV",dV_val,sizeof(float)*nz*nx);
  srite("GR",dR_val,sizeof(float)*nz*nx);  
  //srite("wave",u1_val,sizeof(float)*nz*nx*nt); 
  float *dat = new float[nt*ns*nx];
  Run->GetData(dat);
  srite("data",dat,sizeof(float)*nt*nx*ns);
  //return 0; 
  */
  
  cout << "========================== Conjugate Gradient ==========================" <<endl; 
  double objf, curobjf;              
  float *modl = new float[2*mblock];
  float *dmodl = new float[2*mblock];
  float *grad = new float[2*mblock];
  float *modlKs  = new float[(niter+1)*mblock];          // m_k as number of iteration
  float *modlRs = new float[(niter+1)*mblock];           // m_r as number of iteration
  float *gradks = new float[(niter+1)*mblock];           // grad_k as number of iteration
  float *gradrs = new float[(niter+1)*mblock];           // grad_r as number of iteration
  float *gradis = new float[(niter+1)*mblock];           // grad_r as number of iteration
  int *nfunc = new int[niter+1];                         // number of function evaluations
  float *objfs= new float[niter+1];                      // objective functions
  
  for(int i=0; i < 2*mblock; i++)cmodl[i]=0.;
  ncg->run(cmodl,grad,&objf,nfunc,objfs, modlKs,modlRs, gradks,gradrs);
  cout << "=============Finished calculation.=============="<<endl;
  for(int i=0;i<niter;i++){
    cout<<nfunc[i] << ", " << objfs[i] << endl;
  } 
  
  
  cout << "========================== Output inversion results ==========================" <<endl;  
  // rotate parameter so that MKs = velocity, MRs = density
  float tmpMR, tmpMV,tmpGR,tmpGV,tmpGI;
  for(int iter=0;iter<(niter+1);iter++){
    int shift = iter*mblock;
    for(int i=0;i<mblock;i++){
      tmpMR = modlRs[shift+i];
      tmpMV = 0.5*modlKs[shift+i]-0.5*modlRs[shift+i];
      modlRs[shift+i] = RM_val[i]*exp(tmpMR);
      modlKs[shift+i] = VM_val[i]*exp(tmpMV);
      tmpGR = gradrs[shift+i];
      tmpGV = 0.5*gradks[shift+i]-0.5*gradrs[shift+i];
      tmpGI = 0.5*gradks[shift+i]+0.5*gradrs[shift+i];
      gradis[shift+i] = tmpGI;
      gradks[shift+i] = tmpGV;   // k for V
      gradrs[shift+i] = tmpGR;
    }
  } 
  srite("Vs",modlKs,sizeof(float)*nz*nx*(niter+1));          // output V model
  srite("Rs",modlRs,sizeof(float)*nz*nx*(niter+1));          // output R model
  srite("gradVs",gradks,sizeof(float)*nz*nx*(niter+1));          // output V model
  srite("gradRs",gradrs,sizeof(float)*nz*nx*(niter+1));          // output R model
  srite("gradIs",gradis,sizeof(float)*nz*nx*(niter+1));          // output R model
  
  // rotate parameter so that MKs = impedance 
  for(int i=0;i<(niter+1)*mblock;i++){
    modlKs[i] = modlKs[i]*modlRs[i];
  }
  srite("Is",modlKs,sizeof(float)*nz*nx*(niter+1));          // output Impedance model

  Run->GetData(dataRes);
  srite("dataFinal",dataRes,sizeof(float)*nt*nx*ns);

  cout << "========================== End of Program, Done! ==========================" <<endl; 
  //srite("gradks",gradks,sizeof(float)*nz*nx*(niter+1));
  //srite("gradrs",gradrs,sizeof(float)*nz*nx*(niter+1));
  //Run->GetData(dat);                                          // 
  //srite("datares",dat,sizeof(float)*nt*nx*ns);
  return 0;  
  

}
