#include <iostream>
#include <cstdlib>
#include <stdbool.h>
#include <vector>
#include <cmath>
#include<string>
#include "seplib.h"

#include <omp.h>
#include "Laplacian.h"
#include "Source.h"
#include "SensorPad.h"
#include "Grad.h"
#include "Interpolation.h"
#include "Born.h"
#include "Prop.h"
#include "Engine.h"
#include "CreateModel.h"
#include "HessInv.h"
#include "Wrapper.h"
#include "NCG.h"
#include "utilities.h"
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
extern "C" {
#include "sep3d.h"
}
using namespace std;
using namespace SEP;



/*
Beta version for Multiparameter FWI

Generated by Yinbin: 2015-07-06

Assume: 
	1) source term
	2) initial condition with gradient = 0
	3) K[GPa], t[ms], rho[g/cm^3]
*/

int main(int argc, char *argv[]) {
  cout << "========================================"<<endl;
  cout << "                                        "<<endl;
  cout << "========================================"<<endl;
  cout << "              Test Modulus.        "<<endl;
  initpar(argc,argv);       // Initialize the sep library
  omp_set_num_threads(1);   // Initialize the parallel environment
  
  // Define dimension of interest
  int nx, nz, nt;             // size of computational grid
  int sx, sz, ns, oxs, dxs;   // source geometry
  int boundary;               // size of zero boudary
  float dx, dz, dt;           // propagation parameter
  int nb, bsize;              // nb is the number of cores, bsize is shot per core
  int sample;                 // data sampling
  int mblock;                 // size of velocity model
  int type;                   // type of inversion, 1 for (V,r=const), 2 for (V=const,I), 10 for joint inversion
  int niter;                  // number of iteration for CG method

   // Get parameters
  getch("nb","d",&nb);              nb = 1;              
  getch("type","d",&type);      
  getch("bsize","d",&bsize);        bsize=1; ns = nb*bsize;
  getch("nx","d",&nx);              
  getch("nz","d",&nz);              mblock=nx*nz;         
  getch("nt","d",&nt);              nt = 800;     
  getch("boundary","d",&boundary);  
  getch("dx","f",&dx);              dz = dx;
  getch("sz","d",&sz);              
  getch("oxs","d",&oxs);            
  getch("dxs","d",&dxs);            
  getch("dt","f",&dt);              
  getch("ns","d",&ns);             
  getch("sample","d",&sample);      
  getch("niter","d",&niter);  
  // Summary of jobs
  cout << "========================== Beginning of FWI for Marmousi model ==========================" <<endl;  
  cout << "Number of cores = "   << nb << endl;
  cout << "Jobs on each core = " << bsize << endl;
  cout << "Number of shots = "   << ns << endl;
  if(type==1)  cout << "Velocity estimation for constant rho" << endl;
  if(type==2)  cout << "Impedance estimation for constant velocity" << endl;
  if(type==10) cout << "Joint parameter estimation" << endl;  
  cout << "nx = " << nx << endl;
  cout << "nz = " << nz << endl;
  cout << "nt = " << nt << endl;
  cout << "boundary = " << boundary << endl;
  cout << "dx = " << dx <<"[m]"<< endl;  
  cout << "sz = " << sz << endl;
  cout << "oxs = " << oxs << endl;
  cout << "dxs = " << dxs << endl;
  cout << "dt = " << dt << "[ms]"<< endl;
  cout << "sample = " << sample << endl;
  cout << "Number of CG steps = " << niter << endl;
  
  // Set up space 
  inCoreFloat1D *sspace = new inCoreFloat1D(nt);               // space for source
  inCoreFloat2D *mspace = new inCoreFloat2D(nz,nx);            // space for velocity, reflectivity
  inCoreFloat2D *bmspace= new inCoreFloat2D(2*nz,nx);          // space for velocity, reflectivity
  inCoreFloat3D *uspace = new inCoreFloat3D(nt,nz,nx);         // space for wavefield
  inCoreFloat2D *dspace = new inCoreFloat2D(nt,nx);            // space for data
 
  
  // Set up vectors
  Vector *shot     = new Vector(sspace);          // shot vector
  Vector *VT       = new Vector(mspace);          // true velocity
  Vector *RT       = new Vector(mspace);          // true density
  Vector *VM       = new Vector(mspace);          // starting velocity
  Vector *RM       = new Vector(mspace);          // starting density
  Vector *u1       = new Vector(uspace);          // wave field
  Vector *MKT      = new Vector(mspace);          // K_inv for the true model
  Vector *MRT      = new Vector(mspace);          // rho_inv for the true model
  Vector *MKM      = new Vector(mspace);          // K_inv for the starting model
  Vector *MRM      = new Vector(mspace);          // rho_inv for the starting model
 
  // Creating pointers
  float *__restrict__ shot_val  = ((dataInCoreFloat*)shot->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VT_val    = ((dataInCoreFloat*)VT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ VM_val    = ((dataInCoreFloat*)VM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RT_val    = ((dataInCoreFloat*)RT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ RM_val    = ((dataInCoreFloat*)RM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKT_val   = ((dataInCoreFloat*)MKT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MKM_val   = ((dataInCoreFloat*)MKM->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRT_val   = ((dataInCoreFloat*)MRT->getDataContainer())->getFloatPtr(); 
  float *__restrict__ MRM_val   = ((dataInCoreFloat*)MRM->getDataContainer())->getFloatPtr();   
  float *__restrict__ u1_val    = ((dataInCoreFloat*)u1->getDataContainer())->getFloatPtr();  
 
  // Read in velocity and density
  cout << "========================== Reading in velocity and density model ==========================" <<endl;  
  sreed("wavelet",shot_val,sizeof(float)*nt); 
  sreed("VT",VT_val,sizeof(float)*nz*nx);
  sreed("VM",VM_val,sizeof(float)*nz*nx);
  sreed("RT",RT_val,sizeof(float)*nz*nx);
  sreed("RM",RM_val,sizeof(float)*nz*nx);

  for(int iz=210;iz < nz; iz++){
    for(int ix=0;ix<nx;ix++){
      int i = iz*nx+ix;
      RT_val[i] = 10.;
      RM_val[i] = 10.;
    }
  }
  for(int i=0;i<mblock;i++){
    MKT_val[i] = 1./(RT_val[i]*VT_val[i]*VT_val[i]);
    MKM_val[i] = 1./(RM_val[i]*VM_val[i]*VM_val[i]);  
    MRT_val[i] = 1./RT_val[i];
    MRM_val[i] = 1./RM_val[i];
  }

  int water =0;
  // creating weighting function
  float *weight = new float[mblock];
  for(int iz=0;iz < nz;iz++){
    double coef = exp(-1.*(iz-70)*(iz-70)/100.);
    //coef = 1.;
    if(iz >= 70) coef = 1.;
    for(int ix=0;ix<nx;ix++){
      int pos = iz*nx+ix;
      weight[pos] = coef;
    }
  }
  //srite("weight",weight,sizeof(float)*nx*nz);

  //Constructing operators
  cout << "========================== Constructing operator ==========================" <<endl; 
  float *cmodl = new float[2*mblock];
  for(int i=0; i < mblock; i++){
    cmodl[i] =MKM_val[i];
    cmodl[i+mblock] =MRM_val[i];
  }   
  std::vector<int> sxv;
  std::vector<int> szv;
  for(int i=0; i < nb*bsize; i++){
    sxv.push_back(60);
    szv.push_back(58);
  }
  Engine *Run = new Engine();
  Run->SetSpace(mspace, dspace, uspace);
  Run->SetShot(shot,sxv,szv);
  Run->SetPropPar(nb,bsize,boundary,dt,dx,dz,sample,water,weight);
  Run->SetTrueModel(MKT,MRT);

  // Collecting data
  cout << "========================== Collecting data ==========================" <<endl; 
  float *dataTrue= new float[nt*nx*ns]; 
  float *dataRes = new float[nt*nx*ns]; 
  Run->GetTrueData(dataTrue);
  Run->SetModel(MKM,MRM);
  Run->CreateData();
  Run->GetFWave(u1,0);
  Run->GetData(dataRes);
  for(int i=0;i<nt*nx*ns;i++) dataRes[i] = dataRes[i] + dataTrue[i];
  srite("wave",u1_val,sizeof(float)*nt*nx*nz);
  srite("dataT",dataTrue,sizeof(float)*nt*nx*ns);
  srite("dataR",dataRes,sizeof(float)*nt*nx*ns);
  cout << "========================== End of Program, Done! ==========================" <<endl; 

  return 0;  
  

}
