#ifndef WRAPPER_H
#define WRAPPER_H 1

#include "base/Operator.h"
#include "base/DataInCoreFloat.h"
#include "base/inCoreFloat.h"
#include "Clean.h"
#include "Engine.h"
#include <cmath>
#include <omp.h>
#include "mkl_rci.h"
#include "mkl_blas.h"
#include "mkl_spblas.h"
#include "mkl_service.h"
typedef long long llong;

/*
  Wrapper function
  Used to control gradient and objective function
  
  First Created: 2015-11-03, Yinbin Ma
  Last Modified: 2015-11-03
  
*/


namespace SEP{

class Wrapper{
  public:
    Wrapper(){}
    virtual void Evaluate(float *modl, float *grad, double *objf)=0;
    //virtual void Update(float *modl, float *grad, double *objf)=0;
    virtual void SetO(float *modl)=0;
    virtual void SetType(int _type)=0; 
    virtual void Hessian(Vector *G1, Vector *G2, Vector *H1, Vector *H2)=0;  
    virtual void HessDot() = 0;
    float *curmodl;
    float *curgrad;

    double objf;
    int mflag, gflag, oflag; // flag for the model update
    int type;
  private:

};

class VRWrapper: public Wrapper{
  // type == 1: (v,r) velocity alone
  // type == 2: (v,I) impedance alone
  // type == 3: (v,I) velocity alone
  // type == 10: joint inversion
  public:
    VRWrapper(Space *mod, int _n, Engine *_Run){
      mspace = (inCoreFloat *)mod;
      Kmodl  = new Vector(mspace);
      Rmodl  = new Vector(mspace);
      Kgrad  = new Vector(mspace);
      Rgrad  = new Vector(mspace);
      Kdmodl = new Vector(mspace);
      Rdmodl = new Vector(mspace);
      n = _n;
      Run = _Run;
      curmodl = new float[n];
      curgrad = new float[n];
      curobjf = 1.e19;
      modl0 = new double[2*n];
      mflag = 0.;
      gflag = 0.;
      oflag = 0.;
      type =10;
    }
    void SetO(float *modl){
      for(int i=0;i<2*n;i++){
        modl0[i] = modl[i];
      }
    }
    void SetType(int _type){type = _type;}
    void Evaluate(float *modl, float *grad, double *objf){
      float *__restrict__ KM_val    = ((dataInCoreFloat*)Kmodl->getDataContainer())->getFloatPtr(); 
      float *__restrict__ RM_val    = ((dataInCoreFloat*)Rmodl->getDataContainer())->getFloatPtr(); 
      float *__restrict__ KG_val    = ((dataInCoreFloat*)Kgrad->getDataContainer())->getFloatPtr();       
      float *__restrict__ RG_val    = ((dataInCoreFloat*)Rgrad->getDataContainer())->getFloatPtr();
      for(int i=0; i < n; i++){
        KM_val[i] = modl0[i]*exp(-1.*modl[i]);            //  m_K
        RM_val[i] = modl0[i+n]*exp(-1.*modl[i+n]);        //  m_rho
      }  
      Run->SetModel(Kmodl,Rmodl);    // because we need to inject K and rho in the engine
      Run->GradientFunc(Kgrad,Rgrad);
      if(type==10){
        for(int i=0; i < n; i++){
          grad[i]   = KG_val[i];
          grad[i+n] = RG_val[i];
        }
      }
      if(type==20){
        for(int i=0; i < n; i++){
          grad[i]   = 5.*KG_val[i]+RG_val[i];
          grad[i+n] = KG_val[i]+RG_val[i];
        }      
      }
      if(type==30){
        for(int i=0; i < n; i++){
          grad[i]   = 2.*KG_val[i];
          grad[i+n] = 2.*RG_val[i];
        }      
      }
      if(type==1){
        // (V, rho) only estimate V
        for(int i=0; i < n; i++){
          grad[i]   = KG_val[i];
          grad[i+n] = 0.;
        }      
      }
      if(type==2){
        // (V, I) only estimate I
        for(int i=0; i < n; i++){
          grad[i]   = 0.5*(KG_val[i]+RG_val[i]);
          grad[i+n] = 0.5*(KG_val[i]+RG_val[i]);
        }      
      }
      if(type==3){
        // (V, I) only estimate V
        for(int i=0; i < n; i++){
          grad[i]   = 0.5*(KG_val[i]-RG_val[i]);
          grad[i+n] = -0.5*(KG_val[i]-RG_val[i]);
        }           
      }        
        *objf = Run->OBJF();
    }
    
    void Hessian(Vector *dM1, Vector *dM2, Vector *HM1, Vector *HM2){
      HM1->zeroElement();
      HM2->zeroElement();
      float *__restrict__ dM1_val    = ((dataInCoreFloat*)dM1->getDataContainer())->getFloatPtr(); 
      float *__restrict__ dM2_val    = ((dataInCoreFloat*)dM2->getDataContainer())->getFloatPtr(); 
      float *__restrict__ HM1_val    = ((dataInCoreFloat*)HM1->getDataContainer())->getFloatPtr();       
      float *__restrict__ HM2_val    = ((dataInCoreFloat*)HM2->getDataContainer())->getFloatPtr();
      float *__restrict__ dK_val     = ((dataInCoreFloat*)Kdmodl->getDataContainer())->getFloatPtr(); 
      float *__restrict__ dR_val     = ((dataInCoreFloat*)Rdmodl->getDataContainer())->getFloatPtr(); 

      if(type==10){      // Joint inversion for (K,rho)
        for(int i=0; i < n; i++){
          dK_val[i]   = dM1_val[i];
          dR_val[i]   = dM2_val[i];
        }
      }
      if(type==20){      // Joint inversion for (V, rho)
        for(int i=0; i < n; i++){
          dK_val[i]   = 2*dM1_val[i]+dM2_val[i];  // get dK
          dR_val[i]   = dM2_val[i];               // get dR
        }      
      }
      if(type==30){      // Joint inversion for (V, I)
        for(int i=0; i < n; i++){
          dK_val[i]   = dM1_val[i]+dM2_val[i];  // get dK
          dR_val[i]   = dM2_val[i]-dM1_val[i];  // get dR
        }      
      }
      Run->GaussHessianFunc(Kdmodl,Rdmodl,HM1,HM2);
      // rotate back
      float temp1, temp2;
      if(type==10){      // Joint inversion for (K,rho)
        // nothing need to be done
      }
      if(type==20){      // Joint inversion for (V, rho)
        for(int i=0; i < n; i++){
          temp1 = HM1_val[i];
          temp2 = HM2_val[i];
          HM1_val[i] = 2.*temp1;
          HM2_val[i] = temp1+temp2;
        }      
      }
      if(type==30){      // Joint inversion for (V, I)
        for(int i=0; i < n; i++){
          temp1 = HM1_val[i];
          temp2 = HM2_val[i];
          HM1_val[i] = temp1-temp2;
          HM2_val[i] = temp1+temp2;
        }        
      }      
    }
  void HessDot(){
    Vector *V1 = new Vector(Kdmodl); V1->random();
    Vector *V2 = new Vector(Kdmodl); V2->random();
    Vector *U1 = new Vector(Kdmodl); U1->random();
    Vector *U2 = new Vector(Kdmodl); U2->random();   
    Kdmodl->zeroElement();
    Rdmodl->zeroElement();
    Hessian(V1,V2,Kdmodl,Rdmodl);
    std::cout << " First =  " << Kdmodl->innerProd(*U1)+Rdmodl->innerProd(*U2) <<std::endl;
    Kdmodl->zeroElement();
    Rdmodl->zeroElement();
    Hessian(U1,U2,Kdmodl,Rdmodl);
    std::cout << " Second =  " << Kdmodl->innerProd(*V1)+Rdmodl->innerProd(*V2) <<std::endl;
  }
  private:
    Engine *Run;
    inCoreFloat *mspace;
    Vector *Kmodl, *Rmodl, *Kgrad, *Rgrad, *Kdmodl, *Rdmodl;
    int n;
    double curobjf;
    double *modl0;
}; 

class VWrapper: public Wrapper{

  public:
    VWrapper(Space *mod, int _n, Engine *_Run){
      mspace = (inCoreFloat *)mod;
      Vmodl  = new Vector(mspace);
      Rmodl  = new Vector(mspace);
      Vgrad  = new Vector(mspace);
      Rgrad  = new Vector(mspace);

      n = _n;
      Run = _Run;
      curmodl = new float[n];
      curgrad = new float[n];
      curobjf = 1.e19;
      mflag = 0.;
      gflag = 0.;
      oflag = 0.;
    }
    void SetO(float *modl){
      for(int i=0;i<2*n;i++){
        modl0[i] = modl[i];
      }
    }
    void Evaluate(float *modl, float *grad, double *objf){
      float *__restrict__ VM_val    = ((dataInCoreFloat*)Vmodl->getDataContainer())->getFloatPtr(); 
      float *__restrict__ RM_val    = ((dataInCoreFloat*)Rmodl->getDataContainer())->getFloatPtr(); 
      float *__restrict__ VG_val    = ((dataInCoreFloat*)Vgrad->getDataContainer())->getFloatPtr();       
      float *__restrict__ RG_val    = ((dataInCoreFloat*)Rgrad->getDataContainer())->getFloatPtr();
      for(int i=0; i < n; i++){
        VM_val[i] = modl[i];
        RM_val[i] = modl[i+n];
      }  
        Run->SetModel(Vmodl,Rmodl); 
        Run->GradientFunc(Vgrad,Rgrad);
     
      for(int i=0; i < n; i++){
        grad[i]   = VG_val[i];
        grad[i+n] = 0.;
      }
        
        *objf = Run->OBJF();
    }
  private:
    Engine *Run;
    inCoreFloat *mspace;
    Vector *Vmodl, *Rmodl, *Vgrad, *Rgrad;
    int n;
    double curobjf;
    double *modl0;
}; 


} // end of namespace
#endif