#include "TimeMarch.h"

namespace SEP {

TimeMarch::TimeMarch(llong nb, float dt, float dx, Vector *Vel){
  init(nb,dt,dx,Vel);
}

void TimeMarch::init(llong _nb, float _dt, float _dx, Vector *Vel){

  // save all propagation parameters
  nb = _nb;
  dt = _dt;
  dx = _dx;
  dtdx2 = (dt/dx)*(dt/dx);
  
  // Set up temporary Vectors
  mspace  = (inCoreFloat *)(Vel->getSpace());
  VTX     = new Vector(Vel);
  rhoInvZ = new Vector(Vel);
  rhoInvX = new Vector(Vel);
  rhoReg  = new Vector(Vel);
  rhoInv  = new Vector(Vel);
  temp    = new Vector(Vel);
  tempX   = new Vector(Vel);
  tempZ   = new Vector(Vel);    
  // Set up Operators
  DX = new GradX(mspace,mspace, nb);
  DZ = new GradZ(mspace,mspace, nb);
  DXT = new GradXT(mspace,mspace, nb);
  DZT = new GradZT(mspace,mspace, nb);
  interpX = new IntX(mspace,mspace);
  interpZ = new IntZ(mspace,mspace);
  Lap     = new Laplacian4(mspace,mspace);

  // Fill in Vectors
  std::vector<long long> in_dims = mspace->get_ns();
  nz = in_dims[0];
  nx = in_dims[1];  
  mblock = nz*nx;
  float *__restrict__ cval = ((dataInCoreFloat*)VTX->getDataContainer())->getFloatPtr();
  for(long long i=0; i<nx*nz;i++){
    cval[i] = cval[i]*dtdx2;
  }     // vtx = (dt/dx)^2 v^2
}

void TimeMarch::SetModel(Vector *Vel, Vector *rho){
  float *__restrict__ vval = ((dataInCoreFloat*)Vel->getDataContainer())->getFloatPtr();
  float *__restrict__ cval = ((dataInCoreFloat*)VTX->getDataContainer())->getFloatPtr();
  float *__restrict__ rval = ((dataInCoreFloat*)rho->getDataContainer())->getFloatPtr();
  float *__restrict__ rreg_val = ((dataInCoreFloat*)rhoReg->getDataContainer())->getFloatPtr();
  float *__restrict__ rinv_val = ((dataInCoreFloat*)rhoInv->getDataContainer())->getFloatPtr();

  for(long long i=0; i<nx*nz;i++){
    cval[i] = vval[i]*vval[i]*dtdx2;
    rreg_val[i] = rval[i];
    rinv_val[i] =   1./rval[i];
  }     // vtx = (dt/dx)^2 v^2  
  interpX->Forward(false,rhoInv,rhoInvX);
  interpZ->Forward(false,rhoInv,rhoInvZ);

}

bool TimeMarch::FMarch(Vector *pre, Vector *cur, Vector *next){
  float *__restrict__ v1 = ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
  float *__restrict__ v2 = ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr(); 
  float *__restrict__ v3 = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
  float *__restrict__ cval = ((dataInCoreFloat*)VTX->getDataContainer())->getFloatPtr();
  float *__restrict__ rreg_val = ((dataInCoreFloat*)rhoReg->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)rhoInvX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)rhoInvZ->getDataContainer())->getFloatPtr();
  float *__restrict__ t = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ tx = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tz = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();

  
  DX->Forward(false,cur,temp);
  
  // ISPC substitution
  ScaleISPC(rinvx_val,t,mblock);
  /*
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvx_val[i];
  }
  */
  DXT->Forward(false,temp,tempX);
  
  // ISPC substitution
  ScaleISPC(rreg_val,tx,mblock);
  /*
  for(llong i=0; i < nx*nz; i++){
    tx[i] = tx[i]*rreg_val[i];
  } 
  */ 
  DZ->Forward(false,cur,temp);
  // ISPC substitution
  ScaleISPC(rinvz_val,t,mblock);
  /*
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvz_val[i];
  }
  */
  DZT->Forward(false,temp,tempZ);
  // ISPC substitution
  ScaleISPC(rreg_val,tz,mblock);
  /*
  for(llong i=0; i < nx*nz; i++){
    tz[i] = tz[i]*rreg_val[i];
  }
  */
  
  // ISPC substitution  
  LeapFrogISPC(cval,v1,tx,tz,v2,v3,mblock);
  /*
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*(tx[i]+tz[i])-v1[i];
  }
  */
  
  /*
  Constant density case
  Lap->Forward(false,cur,temp);
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*t[i]-v1[i];
  } 
  */ 
  return 0;
}

bool TimeMarch::BMarch(Vector *pre, Vector *cur, Vector *next){
  float *__restrict__ v1 = ((dataInCoreFloat*)pre->getDataContainer())->getFloatPtr();
  float *__restrict__ v2 = ((dataInCoreFloat*)cur->getDataContainer())->getFloatPtr(); 
  float *__restrict__ v3 = ((dataInCoreFloat*)next->getDataContainer())->getFloatPtr();
  float *__restrict__ cval = ((dataInCoreFloat*)VTX->getDataContainer())->getFloatPtr();
  float *__restrict__ rreg_val = ((dataInCoreFloat*)rhoReg->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvx_val = ((dataInCoreFloat*)rhoInvX->getDataContainer())->getFloatPtr();
  float *__restrict__ rinvz_val = ((dataInCoreFloat*)rhoInvZ->getDataContainer())->getFloatPtr();
  float *__restrict__ t = ((dataInCoreFloat*)temp->getDataContainer())->getFloatPtr();
  float *__restrict__ tx = ((dataInCoreFloat*)tempX->getDataContainer())->getFloatPtr();
  float *__restrict__ tz = ((dataInCoreFloat*)tempZ->getDataContainer())->getFloatPtr();

  // ISPC substitution
  ProdISPC(rreg_val,v2,tx,mblock);  
  /*
  for(llong i=0; i < nx*nz; i++){
    tx[i] = v2[i]*rreg_val[i];
  }
  */
  DXT->Adjoint(false,temp,tempX);
  // ISPC substitution
  ScaleISPC(rinvx_val,t,mblock);
  /*  
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvx_val[i];
  } 
  */ 
  DX->Adjoint(false,tempX,temp);
  // ISPC substitution
  ProdISPC(rreg_val,v2,tz,mblock);  
  /*
  for(llong i=0; i < nx*nz; i++){
    tz[i] = v2[i]*rreg_val[i];
  }
  */
  DZT->Adjoint(false,temp,tempZ);
  // ISPC substitution
  ScaleISPC(rinvz_val,t,mblock);
  /*  
  for(llong i=0; i < nx*nz; i++){
    t[i] = t[i]*rinvz_val[i];
  } 
  */ 
  DZ->Adjoint(false,tempZ,temp);
  
  // ISPC substitution  
  LeapFrogISPC(cval,v1,tx,tz,v2,v3,mblock);
  /*
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*(tx[i]+tz[i])-v1[i];
  }
  */  
 
  /*
  Lap->Adjoint(false,temp,cur);
  for(llong i = 0; i < nx*nz; i++){
    v3[i] = v2[i]*2.+ cval[i]*t[i]-v1[i];
  } 
  */ 
  return 0;
}


}