#include "CGAlg.h"

/* 
    CGStepper functions
*/


void SEP::CGStepper::run() {     // x is the current model
  // r is the current residue
  // g is the gradient in model space
  // G is the gradient in residue space G = Ag
  double alpha,beta=0.,small=1e-20;  // building array size 10 here? 
  double GG= G.innerProd(G); 
  double GR= G.innerProd(r);

  if(iter==0) alpha = -GR/GG;
  else{
    double SS = S.innerProd(S);
    double GS = G.innerProd(S); 
    double determ = GG*SS*std::max( 1.-(GS/GG)*(GS/SS) , 1e-12 );
    double SR = S.innerProd(r);
    alpha = -(SS*GR - GS*SR)/determ;
    beta  = -(-GS*GR + GG*SR)/determ;
  }
  
  if(verbose){
    std::cout << "iter: " << iter << " , alpha = " << alpha << ", beta = " << beta << "\n";
  }
  s.linearCombo(beta,g,alpha);   //Update solution step alpha*g+beta*s
  S.linearCombo(beta,G,alpha); //Update residual step

  x.linearCombo(1.0,s,1.0);      //Update solution
  r.linearCombo(1.0,S,1.0);    //Update residual
  iter = iter + 1;
}



void SEP::CGRegStepper::run() {     // x is the current model
  // r is the current residue
  // g is the gradient in model space
  // G is the gradient in residue space G = Ag
  double alpha,beta=0.,small=1e-20;  // building array size 10 here? 
  double GG= G1.innerProd(G1) + G2.innerProd(G2); 
  double GR= G1.innerProd(r1) + G2.innerProd(r2);

  if(iter==0) alpha = -GR/GG;
  else{
    double SS = S1.innerProd(S1) + S2.innerProd(S2);
    double GS = G1.innerProd(S1) + G2.innerProd(S2); 
    double determ = GG*SS*std::max( 1.-(GS/GG)*(GS/SS) , 1e-12 );
    double SR = S1.innerProd(r1) + S2.innerProd(r2);
    alpha = -(SS*GR - GS*SR)/determ;
    beta  = -(-GS*GR + GG*SR)/determ;
  }
  
  if(verbose){
    std::cout << "iter: " << iter << " , alpha = " << alpha << ", beta = " << beta << "\n";
  }
  s.linearCombo(beta,g,alpha);   //Update solution step alpha*g+beta*s
  S1.linearCombo(beta,G1,alpha); //Update residual step
  S2.linearCombo(beta,G2,alpha); //Update residual step

  x.linearCombo(1.0,s,1.0);      //Update solution
  r1.linearCombo(1.0,S1,1.0);    //Update residual
  r2.linearCombo(1.0,S2,1.0);    //Update residual
  iter = iter + 1;
}



/*
  CG least square functions
*/

bool SEP::CGAlg::query() {
  if (rnorm < rtol) return true;
  if (nrnorm < nrtol) return true;
  if (iter >= maxiter) return true;
  return false;
} 

void SEP::CGAlg::disp() {
  if(data_res !=NULL) 
    data_res[iter] = rnorm;
  if(verbose) {
    std::cout<<"Iter = " << iter << ", rnorm = "<<rnorm << ", nrnorm = " << nrnorm << std::endl;
  }
}

void SEP::CGAlg::run() {
  // Build a CG Step class
  CGStepper step(x, r, g, G, false);
  rnorm = std::sqrt(r.innerProd(r));
  nrnorm = 1.;
  disp();
  while (!query()) {
    iter = iter + 1;
    A.Adjoint(false,&g,&r); // compute the gradient in model space
    A.Forward(false,&g,&G);   // compute the gradient in data  space
    step.run();
    rnorm = std::sqrt(r.innerProd(r));
    nrnorm = std::sqrt(g.innerProd(g));
    disp();
  } 
}



bool SEP::CGRegAlg::query() {
  if (rnorm < rtol) return true;
  if (nrnorm < nrtol) return true;
  if (iter >= maxiter) return true;
  return false;
} 

void SEP::CGRegAlg::disp() {
  if(data_res !=NULL) 
    data_res[iter] = rnorm;
  if(verbose) {
    std::cout<<"Iter = " << iter << ", rnorm = "<<rnorm << ", nrnorm = " << nrnorm << std::endl;
  }
}

void SEP::CGRegAlg::run() {
  // Build a CG Step class
  CGRegStepper step(x, r1, r2, g, G1, G2,false);
  rnorm = std::sqrt(r1.innerProd(r1) + r2.innerProd(r2));
  nrnorm = 1.;
  disp();
  while (!query()) {
    iter = iter + 1;
    A1.Adjoint(false,&g,&r1); // compute the gradient in model space
    A2.Adjoint(true,&g,&r2);
    A1.Forward(false,&g,&G1);   // compute the gradient in data  space
    A2.Forward(false,&g,&G2);
    step.run();
    rnorm = std::sqrt(r1.innerProd(r1)); // + r2.innerProd(r2));
    nrnorm = std::sqrt(g.innerProd(g));

    disp();
  } 
}



