/*
  Used to join operators in different form.
  We should try to avoid use explicit inCoreFloat, etc.
  Try to complete the functions with high level syntax. 

  I do not like the way I handle RowOP and ColOP. I need additional vector copy, and we really want to avoid it.
  1) We should first fix the usage of Multi-datacontainer, Multi-Space, Multi-Vector.
  2) Then we can fix the problem here, and standardize the Multi-Operator.
  3) Then we would be able to write general solvers. 

  Usage:

  1) d = L' m, transpose operator
     TransOP OP(L);	// initializer
     OP.Forward(add,m,d); //apply operator 

  2) d = alpha m, scaling operator
     scaleOP OP(Space *mod, Space *data, float alpha=1)  // initializer
     OP.setscale(alpha);                      // set scaling factor
     OP.Forward(add,m,d);                 // apply operator

  3) d = L2*L1 m      chain operator
     ChainOP OP(L1, L2);    // initializer 
     OP.Forward(add, m ,d);  // apply operator


  3) d=[d1; d2]= [L1;L2]m   Column operator
      ColOP OP(L1,L2,dspace);     // initializer 
      OP.Forward(add, m, d);      // apply operator



  4) d = [L1, L2] [m1; m2] Row Operator
     SHould have similar implementation as column operator.
     We do not have it yet.

  5)d = [d1;d2] = [L11, 0; 0, L22][m1;m2] Diagonal Operator
     Again, should be simple once we fix column operator.
     We do not have it yet.


*/

#ifndef MULTIOPERATOR_H
#define MULTIOPERATOR_H 1
#include "Space.h"
#include "DataInCoreFloat.h"
#include "inCoreFloat.h"
#include "Operator.h"
#include "Vector.h"
#include <iostream>
typedef long long llong;



namespace SEP {

class TransOP: public Operator{

  public:
    TransOP(Operator &_L)
        :L(_L) {
      inCoreFloat *mod = (inCoreFloat *)(L.getRange());
      inCoreFloat *dat = (inCoreFloat *)(L.getDomain());
      setDomain(mod);
      setRange(dat);
    }
    virtual bool Forward(bool add, Vector *model, Vector *data) {
      L.Adjoint(add,data,model);
      return true;
    }
    virtual bool Adjoint(bool add, Vector *model, Vector *data) {
      L.Forward(add,data,model);
      return true;
    }

  private:
    Operator &L;
};


class ScaleOP: public Operator{

  public:
    ScaleOP(Space *mod, Space *data, float alpha = 1.0) {
      setDomain(mod); 
      setRange(data);
      this->alpha = alpha;
    }

    void SetAlpha(float _alpha) {alpha  = _alpha;}

    virtual bool Forward(bool add, Vector *model, Vector *data) {
      if(add)
        data->linearCombo(1.,model,alpha);
      else
        data->linearCombo(0.,model,alpha);
      return true;
    }
    virtual bool Adjoint(bool add, Vector *model, Vector *data) {
      if(add)
        model->linearCombo(1.,data,alpha);
      else
        model->linearCombo(0.,data,alpha);
      return true;
    }

  private:
    float alpha;

};






class ChainOP: public Operator{

  public:
    ChainOP(Operator &_L1,Operator &_L2)
        :L1(_L1),L2(_L2) {
      inCoreFloat *mod = (inCoreFloat *)(L1.getDomain());
      inCoreFloat *dat = (inCoreFloat *)(L2.getRange());
      inCoreFloat * midspace = (inCoreFloat *)(L1.getRange());
      setDomain(mod); 
      setRange(dat);
      middle = new Vector(midspace);
    }
    virtual bool Forward(bool add, Vector *model, Vector *data) {
      L1.Forward(false,model,middle);
      L2.Forward(add,middle,data);
      return true;
    }
    virtual bool Adjoint(bool add, Vector *model, Vector *data) {
      L2.Adjoint(false,middle,data);
      L1.Adjoint(add,model,middle);
      return true;
    }
    ~ChainOP() {delete middle; middle = NULL;}
  private:
    Operator &L1;
    Operator &L2;
    Vector *middle;
};

class ColOP: public Operator{
  public:
    ColOP(Operator &_L1, Operator &_L2, Vector *data)
        :L1(_L1),L2(_L2) {
      ran1 = (inCoreFloat *)L1.getRange();
      ran2 = (inCoreFloat *)L2.getRange(); 
      setDomain(L1.getDomain());
      setRange((MultiSpace *)data->getSpace());
    }
    virtual bool Forward(bool add, Vector *model, Vector *data) {
      if(!add) data->zeroElement();
      Vector d1(ran1);d1.zeroElement();
      Vector d2(ran2);d2.zeroElement();
      L1.Forward(true,model,&d1);
      L2.Forward(true,model,&d2);
      Vector *d = new MultiVector(&d1,&d2,(MultiSpace *)(data->getSpace()));

      data->linearCombo(1.,*d,1.);
      return true;
    }
    virtual bool Adjoint(bool add, Vector *model, Vector *data) {
      if(!add) model->zeroElement();
      Vector d1(ran1);
      Vector d2(ran2);
      ran1->linearCombo(0.,d1.getDataContainer(),1.,((MultiVector *)data)->getDataContainer1());
      ran2->linearCombo(0.,d2.getDataContainer(),1.,((MultiVector *)data)->getDataContainer2());
      L1.Adjoint(true,model,&d1);
      L2.Adjoint(true,model,&d2);
      return true;
    }

bool DotTest(bool verb/*, Space *dom, Space *ran*/){
  /// Test that (Fx,y) = x*F*y = (x,Fy) 
  /// (Linear algebra identity)

  Vector *v1 = new Vector(((MultiSpace *)range)->getSpace1());
  Vector *v2 = new Vector(((MultiSpace *)range)->getSpace2());
  Vector *m1 = new Vector(domain);//=new Vector(domain);
  MultiVector *d1 = new MultiVector(v1,v2,(MultiSpace *)range);//=new Vector(range);
  Vector *m2 = new Vector(domain);//=new Vector(domain);
  MultiVector *d2 = new MultiVector(v1,v2,(MultiSpace *)range);//=new Vector(range);

  m1->random();
  d2->random();

  Forward(false,m1,d1);
  double dot1=d1->innerProd(*d2);

  Adjoint(false,m2,d2);
  double dot2=m2->innerProd(*m1);

  if(verb) std::cout<<"Dot product test with add=false: \n"<<dot1<<"\n"<<dot2<<"\n";

  Forward(true,m1,d1);
  double dot3=d1->innerProd(*d2);

  Adjoint(true,m2,d2);
  double dot4=m2->innerProd(*m1);

  if(verb) std::cout<<"Dot product test with add=true: \n"<<dot3<<"\n"<<dot4<<"\n";

  return true;
}

  private:
    Operator &L1;
    Operator &L2;
    inCoreFloat *ran1;
    inCoreFloat *ran2;
};








}

#endif