#include "Space.h"
#include "dataContainer.h"

void SEP::MultiSpace::scale(float a, dataContainer  * x, dataContainer * ax) const{
  /// Scale x by a and write it into ax
  ///
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  SEP::MultiDataContainer *ax2=(SEP::MultiDataContainer*) ax;
  sp1->scale(a,x2->getData1(),ax2->getData1());
  sp2->scale(a,x2->getData2(),ax2->getData2());
}

void SEP::MultiSpace::scale(float a, dataContainer * x) const{
  /// Scale x by a and overwrite x with a*x
  ///
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  sp1->scale(a,x2->getData1());
  sp2->scale(a,x2->getData2());
}

// DN use dynamic_cast for RTTI checking?
void SEP::MultiSpace::addInv(dataContainer  * x) const{
  /// Overwrite x with additive inverse 
  /// x = -1*x
  SEP::MultiDataContainer *x2= dynamic_cast<SEP::MultiDataContainer*>( x );
  
  sp1->addInv(x2->getData1());
  sp2->addInv(x2->getData2());
  
}

void SEP::MultiSpace::addInv(dataContainer  * x, dataContainer * xNeg) const{
  /// Write additive inverse of x, -1*x, into xNeg
  /// 
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  SEP::MultiDataContainer *xNeg2=(SEP::MultiDataContainer*) xNeg;
  sp1->addInv(x2->getData1(),xNeg2->getData1());
  sp2->addInv(x2->getData2(),xNeg2->getData2());
}

/* void SEP::MultiSpace::norm(int p, dataContainer * x) const{
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  sp1->norm(p,x2->getData1());
  sp2->norm(p,x2->getData2());
} */

bool SEP::MultiSpace::isCompatible(SEP::Space *otherSpace, bool makeSame) const{
  /// Check whether this space and otherSpace are compatible, either 
  /// because they are the same space or by a deep check. 
  /// By default, compatible spaces are not given the same key, but
  /// if makeSame = true, the key of otherSpace is set to the key of this
  /// space.
  SEP::MultiSpace *s2=(SEP::MultiSpace*) otherSpace;
  if(sp1->isCompatible(s2->getSpace1()) && sp2->isCompatible(s2->getSpace2())){
    if(makeSame) s2->set_key(get_key());
    return true;
  } 
  return false;
}

bool SEP::MultiSpace::isCompatible(SEP::dataContainer *x) const{
  /// Check if a data container, x, is compatible with this space.
  /// 
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;

  if(sp1->isCompatible(x2->getData1())&&
   sp2->isCompatible(x2->getData2())) return true;

  return false; 
}


double SEP::MultiSpace::innerProd(dataContainer * x, dataContainer * y)const{
  /// Take the inner product (sum of element wise products)
  /// of two data containers compatible with the same space
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  SEP::MultiDataContainer *y2=(SEP::MultiDataContainer*) y;
  double sum=0.;
  sum += sp1->innerProd(x2->getData1(),y2->getData1());
  sum += sp2->innerProd(x2->getData2(),y2->getData2());
  
  return sum;
}

void SEP::MultiSpace::linearCombo(float a, dataContainer  * x, float b, dataContainer * y) const{
  /// Calculate x = a*x + b*y
  /// 
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  SEP::MultiDataContainer *y2=(SEP::MultiDataContainer*) y;
  sp1->linearCombo(a,x2->getData1(),b,y2->getData1());
  sp2->linearCombo(a,x2->getData2(),b,y2->getData2());
}


void SEP::MultiSpace::zeroElement(dataContainer * x)const{
  /// Zero out any vector in the MultiSpace by zeroing
  /// it out in the two spaces making up the multispace.
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  sp1->zeroElement(x2->getData1());
  sp2->zeroElement(x2->getData2());
}

/*SEP::MultiSpace SEP::MultiSpace::clone_space()const{
   MultiSpace x;
   x.sp1=sp1;
   x.sp2=sp2;
   return x;
}*/

SEP::MultiDataContainer *SEP::MultiSpace::buildDataContainer()const{
  /// Build a data container that is compatible with the multispace
  /// by building a multi data container.
  dataContainer *d1=sp1->buildDataContainer();
  dataContainer *d2=sp2->buildDataContainer();
  SEP::MultiDataContainer *mdc = new SEP::MultiDataContainer(d1,d2);
  return mdc;
}

SEP::MultiSpace::MultiSpace(Space *s1, Space *s2){
  /// Create a MultiSpace with pointers to the two input spaces.
  /// 
  sp1=s1;
  sp2=s2;
}

long long SEP::MultiSpace::get_n123(){
  /// Return the total dimension of the multi-space assuming no 
  /// overlap (so n123 is sum of n123 for spaces making up multispace)
  long long n1 = sp1->get_n123();
  long long n2 = sp2->get_n123();
  return n1+n2;
}

void SEP::MultiSpace::random(dataContainer * x) const{
  /// Fill a data container x with random numbers.
  ///
  SEP::MultiDataContainer *x2=(SEP::MultiDataContainer*) x;
  sp1->random(x2->getData1());
  sp2->random(x2->getData2());
}
		
