#include "kirchhoff.h"
#include <iostream>
#include <cstdlib>
#include <cmath> 
typedef long long llong;

bool SEP::Kirchfast2d::Forward(bool add, Vector *model, Vector *data){
  /// Forward operator for spray and sum
  ///
  if(!add) data->zeroElement();  // add=False we will erase the data

  float *__restrict__ model_val = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ data_val = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ vrms_val = ((dataInCoreFloat*)Vrms->getDataContainer())->getFloatPtr();

  for(llong ih = -nx + 1; ih < nx; ih++) {
    float h = dx*(float)ih;      // b is offset
    for(llong iz = 1; iz < nt; iz++) {
      float z = t0 + dt * iz; // z = travel-time depth
      float t = std::sqrt(z*z + 4.*h*h/(vrms_val[iz]*vrms_val[iz]));  // fill in later
      llong it = (llong)(0.5 + (t - t0)/dt);
      if (it >= nt) break;
      if(ih != 0) {
        double angle = std::acos(z/t); angle = angle*180/3.1415926535;
        if (angle > amax) continue;
      }
      float amp = (z / t) * std::sqrt(nt*dt / t);
      llong xstart = std::max((llong)0, -ih);
      llong xend = std::min(nx, nx - ih);
      for(llong ix = xstart; ix < xend; ix++) {
        data_val[it + (ix+ih) * nz]  += model_val[iz + ix * nz] * amp;
      } 
    }
  }

  return true;
}

bool SEP::Kirchfast2d::Adjoint(bool add, Vector *model, Vector *data) {
  /// For Adjoint operator, we just need to call Forward operator
  ///
  if(!add) model->zeroElement();  // add=False we will erase the data

  float *__restrict__ model_val = ((dataInCoreFloat*)model->getDataContainer())->getFloatPtr();
  float *__restrict__ data_val = ((dataInCoreFloat*)data->getDataContainer())->getFloatPtr();
  float *__restrict__ vrms_val = ((dataInCoreFloat*)Vrms->getDataContainer())->getFloatPtr();

  for(llong ih = -nx + 1; ih < nx; ih++) {
    float h = dx*(float)ih;      // b is offset
    for(llong iz = 1; iz < nt; iz++) {
      float z = t0 + dt * iz; // z = travel-time depth
      float t = std::sqrt(z*z + 4.*h*h/(vrms_val[iz]*vrms_val[iz]));  // fill in later
      llong it = (llong)(0.5 + (t - t0)/dt);
      if (it >= nt) break;
      if(ih !=0) {
        double angle = std::acos(z/t); angle = angle*180/3.1415926535;
        if (angle > amax) continue;
      }
      float amp = (z / t) * std::sqrt(nt*dt / t);
      llong xstart = std::max((llong)0, -ih);
      llong xend = std::min(nx, nx - ih);
      for(llong ix = xstart; ix < xend; ix++) {
        model_val[iz + ix * nz] += data_val[it + (ix+ih) * nz] * amp;
      } 
    }
  }
  return true;
}


void SEP::Kirchfast2d::init() {
  // Parameters that we could get from space
  inCoreFloat *model_space = (inCoreFloat *)domain;
  inCoreFloat *data_space = (inCoreFloat *)range;
  std::vector<long long> model_dims = model_space->get_ns();
  std::vector<long long> data_dims = data_space->get_ns();
  nz = model_dims[1]; // may change
  nx = model_dims[0]; // may change
  nt = data_dims[1];  // may change the location
  ny = data_dims[0];  // may change

  // default value to make code working
  dz = 1.;
  dx = 1.;
  dt = 1.;
  dy = 1.;
  z0 = 0.;
  x0 = 0.;
  t0 = 0.;
  y0 = 0.;
  amin = 0.;
  amax = 90.;

  // default Vrms
  inCoreFloat1D Vspace(nz);
  Vrms = new Vector(&Vspace);
  float *__restrict__ vrms_val = ((dataInCoreFloat*)Vrms->getDataContainer())->getFloatPtr();
  for (llong i = 0; i < nz; i++)
    vrms_val[i] = 2000.;
}

void SEP::Kirchfast2d::SetModelPar(float z0,float dz,float x0,float dx) {
  this->z0 = z0;
  this->dz = dz;
  this->x0 = x0;
  this->dx = dx;
}

void SEP::Kirchfast2d::SetDataPar(float t0,float dt,float y0,float dy) {
  this->t0 = t0;
  this->dt = dt;
  this->y0 = y0;
  this->dy = dy;
}

void SEP::Kirchfast2d::SetAperture(float amin, float amax) {
  this->amin = amin;
  this->amax = amax;
}

void SEP::Kirchfast2d::clean() {
   Vrms = NULL;
  //delete Vrms;
  //Vrms = NULL;
}
