/***
 *  These C++ classes in flexSIMD.h enable performance and allow easy switching 
 *  between either SSE, AVX, FMA3 or FMA4.  
 *  The file flexSIMD.cpp  has sample code that demonstrates some approaches for how to get speed 
 *  from FD kernels using these classes.  
 *  These classes are simpler than Intel's F32vec4 and F32vec8, and have no license restrictions.   
 *
 *  The kernels should achieve 8-15 GFlops on most modern cores in SSE mode.  Unfortunately, I found no 
 *  advantage to using AVX or FMA on recent CPUs.  This is very, very disappointing.  I have performance 
 *  test results of this code on numerous CPUs in a an accompanying file.
 *
 *  Perhaps this code can be used as a basis for experts to show how to achieve better performance from new
 *  hardware, especially heterogeneous hardware.  
 *
 * These classes easily switch between either SSE, AVX, FMA3 or FMA4 float
 * operations for multiple & add.  The need for efficiency strongly affects the class design.  
 * We are only using language features that are very predictable by the compiler, just a little 
 * more than macros.  This is why we can't use virtual functions and why they are fixed length.
 * We also seek to avoid temporary return variables from functions.
 * We could have used templates for these classes, but I don't trust templates.
 *
 * Note that SSE & AVX is most efficient when they operate on several groups at once.
 * A multiply may take 3-5 operations to complete, but
 * the CPU can work on 3-5 at the same time, IF THEY ARE INDEPENDENT OF EACH OTHER.  We can't work on too
 * many floating point values, because then we can't keep most data in the 16 ymm (AVX) registers.
 * I suggest you operate on 32 float values at once.  The reason these classes only contain 16 floats
 * rather than 32 is that the SSE class would have too many _mm128 objects which would be inefficient
 * usage of registers.  You should use 2 of these SIMD_CLASS16 in sequence.   See example code at end.  Note that
 * the compiler can reorder instructions to reduce waiting for instruction dependency.  
 *
 * Some CPUs are moving to 16 float values (512 bits) SIMD instructions.
 * And 16 float values (64 bytes) is the cache line size used by most (all?) current CPUs.
 *
 * 
 *
 *  This code was created by Christof Stork.  It is in the public domain, freely available, with no license.
 *  You are free to copy and modify the code.  If you use or copy much of the code, I request that you credit me.
 *  If you make improvements, I'm interested in what you did.  cstork@pcisys.net
 *  Thank you!  

 */
#define BLOCK_SIZE 16
#include <assert.h>
//#include <flexSIMD.h>


#ifndef __SSE__
#error  You have an ancient non-SSE compiler, or you need to add compiler flags.
#endif


#ifdef  __GNUC__
#warning  flag __GNUC__ found.  Using Gnu compiler includes.
#include <xmmintrin.h>
//#include <x86intrin.h>
#ifdef __FMA4__
#include <fma4intrin.h>
#endif

#else  // not GNUC
#warning  flag __GNUC__ not found.  Using Intel compiler includes.

#include <xmmintrin.h>

#ifdef __AVX__
#include <immintrin.h>
#endif  // AVX

#ifdef __FMA__
#define  __FMA3__
#include <fmaintrin.h>
#endif


#endif // GNUC



#ifdef __AVX__
#warning	AVX is supported by your compiler!
#else
#warning	AVX is not supported by your compiler, or you need to add compiler flags!
#endif
#ifdef __FMA4__
#warning	Compiler supports FMA4!
#endif
#if  defined ( __FMA__ ) ||  defined ( __FMA3__ )
#warning	Compiler supports FMA3!
#endif


#ifdef __FMA4__
#define		SIMD_Float16	AVX_FMA4_Float16
#elif defined ( __FMA__ ) ||  defined ( __FMA3__ )
#define		SIMD_Float16	AVX_FMA3_Float16
#elif defined __AVX__
#define		SIMD_Float16	AVX_Float16
#else
#define		SIMD_Float16	SSE_Float16
#endif


// This class hides the _mm128 & _mm256 objects.
// This class is needed so the same code work for SSE or AVX  SIMD_Float16
class SIMD_constant {
public:
union {
#ifdef __AVX__
	__m256  _ymm;
#endif
	__m128  _xmm;
}; 
};



// This is just an interface class to show how all the other classes work.
// We can't use virtual functions because that kills performance
// We don't bother implementing any of the functions since they won't be used.
class  Abstract_Float16  {

	static const char*  description() {return "Abstract class for 16 floats. ";}

private:
	// private to make sure no one creates one of these.
	Abstract_Float16()  // empty constructor
	{
	}
	Abstract_Float16( float* array);  //initialize with array of 16 floats.
	// array needs to be aligned to 16 or 32 bytes.

	Abstract_Float16(Abstract_Float16& dum)  // copy constructor
	{
	}
	Abstract_Float16(float c)  // default all values to a constant, such as 0.0
	{
	}

public:

	inline void load( float* array); // read in 16 float values
	inline void loadu( float* array); // read in 16 float values, unaliagned

	inline void stream( float* array);  // Efficient write bypassing cache.  Memory is not read into cache which
	// the cache has to do to "own the memory unit.
	// Use this if you don't need the output for a long time.

	inline void store( float* array);  // write saving output in cache.  Causes a memory read so the cache owns the memory.


	// These operators are easier to implement because they don't require a temporary.
	inline Abstract_Float16& operator*=( const Abstract_Float16& B);
	inline Abstract_Float16& operator*=( const float* Barray);  // Use directly with aligned array of 16 floats
	inline Abstract_Float16& operator*=( const SIMD_constant& sConst  ); // use a constant for all operations.

	inline Abstract_Float16& operator+=( const Abstract_Float16& B);
	inline Abstract_Float16& operator+=( const float* B);
	inline Abstract_Float16& operator+=( const SIMD_constant& sConst  );

	inline Abstract_Float16& operator-=( const Abstract_Float16& B);
	inline Abstract_Float16& operator-=( const float* B);
	inline Abstract_Float16& operator-=( const SIMD_constant& sConst  );

	// We are only bothering with multiply, add, and subtract in this class.  You can expand to the other functions.

	// we can't use the operator* because that creates an inefficient temporary variable
	// Use like this:
	//     Abstract_Float16  r1, A, B;
	//     r1.mult( A, B );    //  r1= A * B
	//     r1.add( A, B );     //  r1= A + B
	//     r1.sub( A, B );     //  r1= A - B
	inline Abstract_Float16& mult(const Abstract_Float16& A, const Abstract_Float16& B);
	inline Abstract_Float16& mult(const Abstract_Float16& A, const SIMD_constant& sConst );
	inline Abstract_Float16& mult(const Abstract_Float16& A, const float* B);

	inline Abstract_Float16& add(const Abstract_Float16& A, const Abstract_Float16& B);
	inline Abstract_Float16& add(const Abstract_Float16& A, const SIMD_constant& sConst );
	inline Abstract_Float16& add(const Abstract_Float16& A, const float* B);

	inline Abstract_Float16& sub(const Abstract_Float16& A, const Abstract_Float16& B);
	inline Abstract_Float16& sub(const Abstract_Float16& A, const SIMD_constant& sConst );
	inline Abstract_Float16& sub(const Abstract_Float16& A, const float* B);

	// new CPUs support efficient multiply-add instructions.
	// Use like this:
	//     Abstract_Float16  r1; 
	//     r1.fMultAdd( B, C );    //  r1= r1 + B * C
	inline Abstract_Float16& fMultAdd( const Abstract_Float16& mult1, const Abstract_Float16& mult2 );
	inline Abstract_Float16& fMultAdd( const Abstract_Float16& mult1,  const SIMD_constant& mult2Const );
};



class SSE_Float16   {
	// perform basic SSE instructions on 16 floating point values

	//  most recent CPUs have 16 xmm registers

	__m128 _xmm0;
	__m128 _xmm1;
	__m128 _xmm2;
	__m128 _xmm3;


public:

	static  const char*  description() {return "Basic SSE, 16 floats";}

	SSE_Float16()  // empty constructor
	{
	}
	SSE_Float16( float* array) __attribute__((always_inline))  //initialize to float array
													{
		assert( ((unsigned long)array) % 16 == 0 );
		_xmm0= _mm_load_ps( array );
		_xmm1= _mm_load_ps( array+ 4 );
		_xmm2= _mm_load_ps( array+ 2*4 );
		_xmm3= _mm_load_ps( array+ 3*4 );
													}
	SSE_Float16( const SSE_Float16& B)  // copy constructor
	{
		_xmm0=  B._xmm0 ;
		_xmm1=  B._xmm1 ;
		_xmm2=  B._xmm2 ;
		_xmm3=  B._xmm3 ;
	}
	SSE_Float16( float constVal )  // copy constructor
	{
		_xmm0= _mm_set1_ps( constVal );
		_xmm1= _xmm0;
		_xmm2= _xmm0;
		_xmm3= _xmm0;
	}

	inline void static setConstant( SIMD_constant& sConst, const float& constVal )  __attribute__((always_inline))  {
		sConst._xmm= _mm_set1_ps( constVal );
	}

	inline void stream( float* array) __attribute__((always_inline))   // write bypassing cache
													{
		assert( ((unsigned long)array) % 16 == 0 );
		_mm_stream_ps( array,     _xmm0 );
		_mm_stream_ps( array+ 4,   _xmm1 );
		_mm_stream_ps( array+ 2*4, _xmm2 );
		_mm_stream_ps( array+ 3*4, _xmm3 );
													}
	inline void store( float* array)  // write including cache
	{
		assert( ((unsigned long)array) % 16 == 0 );
		_mm_store_ps( array,     _xmm0 );
		_mm_store_ps( array+ 4,   _xmm1 );
		_mm_store_ps( array+ 2*4, _xmm2 );
		_mm_store_ps( array+ 3*4, _xmm3 );
	}
	inline void load( float* array)  //
	{
		assert( ((unsigned long)array) % 16 == 0 );
		_xmm0= _mm_load_ps( array      );
		_xmm1= _mm_load_ps( array+ 4    );
		_xmm2= _mm_load_ps( array+ 2*4  );
		_xmm3= _mm_load_ps( array+ 3*4  );
	}
	inline void loadu( float* array)  //
	{
		_xmm0= _mm_loadu_ps( array      );
		_xmm1= _mm_loadu_ps( array+ 4    );
		_xmm2= _mm_loadu_ps( array+ 2*4  );
		_xmm3= _mm_loadu_ps( array+ 3*4  );
	}

	// The reason that these don't call  the operator* is that the compiler implements
	// a = _mm_mul_ps(b,c) by  doing this:
	//       a=b; a=_mm_mul_ps(a,c);  _mm_mul_ps() is really the *= operator.
	//       _mm_mul_ps() needs to output to one of the input arguments.
	// If we implement *= by calling operator*, I'm not sure the compiler is smart enough to see the copy isn't needed.
	inline SSE_Float16& operator*=( const SSE_Float16& B)  __attribute__((always_inline))
	{
		_xmm0= _mm_mul_ps( _xmm0, B._xmm0 );
		_xmm1= _mm_mul_ps( _xmm1, B._xmm1 );
		_xmm2= _mm_mul_ps( _xmm2, B._xmm2 );
		_xmm3= _mm_mul_ps( _xmm3, B._xmm3 );

		return *this;
	}

	inline SSE_Float16& operator*=( const float* B)
	{
		assert( ((unsigned long)B) % 16 == 0 );
		_xmm0= _mm_mul_ps( _xmm0, _mm_load_ps(B+0*4) );
		_xmm1= _mm_mul_ps( _xmm1, _mm_load_ps(B+1*4) );
		_xmm2= _mm_mul_ps( _xmm2, _mm_load_ps(B+2*4) );
		_xmm3= _mm_mul_ps( _xmm3, _mm_load_ps(B+3*4) );

		return *this;
	}
	inline SSE_Float16& operator*=( const SIMD_constant& sConst )
	{
		_xmm0= _mm_mul_ps( _xmm0, sConst._xmm );
		_xmm1= _mm_mul_ps( _xmm1, sConst._xmm );
		_xmm2= _mm_mul_ps( _xmm2, sConst._xmm );
		_xmm3= _mm_mul_ps( _xmm3, sConst._xmm );

		return *this;
	}

	inline SSE_Float16& operator+=( const SSE_Float16& B)
	{
		_xmm0= _mm_add_ps( _xmm0, B._xmm0 );
		_xmm1= _mm_add_ps( _xmm1, B._xmm1 );
		_xmm2= _mm_add_ps( _xmm2, B._xmm2 );
		_xmm3= _mm_add_ps( _xmm3, B._xmm3 );

		return *this;
	}
	inline SSE_Float16& operator+=( const float* B)  __attribute__((always_inline))
	{
		assert( ((unsigned long)B) % 16 == 0 );
		_xmm0= _mm_add_ps( _xmm0, _mm_load_ps(B+0*4) );
		_xmm1= _mm_add_ps( _xmm1, _mm_load_ps(B+1*4) );
		_xmm2= _mm_add_ps( _xmm2, _mm_load_ps(B+2*4) );
		_xmm3= _mm_add_ps( _xmm3, _mm_load_ps(B+3*4) );

		return *this;
	}
	inline SSE_Float16& operator+=( const SIMD_constant& sConst )
	{
		_xmm0= _mm_add_ps( _xmm0, sConst._xmm );
		_xmm1= _mm_add_ps( _xmm1, sConst._xmm );
		_xmm2= _mm_add_ps( _xmm2, sConst._xmm );
		_xmm3= _mm_add_ps( _xmm3, sConst._xmm );

		return *this;
	}

	inline SSE_Float16& operator-=( const SSE_Float16& B)
	{
		_xmm0= _mm_sub_ps( _xmm0, B._xmm0 );
		_xmm1= _mm_sub_ps( _xmm1, B._xmm1 );
		_xmm2= _mm_sub_ps( _xmm2, B._xmm2 );
		_xmm3= _mm_sub_ps( _xmm3, B._xmm3 );

		return *this;
	}
	inline SSE_Float16& operator-=( const float* B)
	{
		assert( ((unsigned long)B) % 16 == 0 );
		_xmm0= _mm_sub_ps( _xmm0, _mm_load_ps(B+0*4) );
		_xmm1= _mm_sub_ps( _xmm1, _mm_load_ps(B+1*4) );
		_xmm2= _mm_sub_ps( _xmm2, _mm_load_ps(B+2*4) );
		_xmm3= _mm_sub_ps( _xmm3, _mm_load_ps(B+3*4) );

		return *this;
	}
	inline SSE_Float16& operator-=( const SIMD_constant& sConst  )
	{
		_xmm0= _mm_sub_ps( _xmm0, sConst._xmm );
		_xmm1= _mm_sub_ps( _xmm1, sConst._xmm );
		_xmm2= _mm_sub_ps( _xmm2, sConst._xmm );
		_xmm3= _mm_sub_ps( _xmm3, sConst._xmm );

		return *this;
	}


	//  SSE_Float16  lhs.mult( A, B);   // lhs= A*B
	// we can't use the operator* because that creates an inefficient temporary variable
	inline SSE_Float16& mult(const SSE_Float16& A, const SSE_Float16& B)
	{
		_xmm0= _mm_mul_ps( A._xmm0, B._xmm0 );
		_xmm1= _mm_mul_ps( A._xmm1, B._xmm1 );
		_xmm2= _mm_mul_ps( A._xmm2, B._xmm2 );
		_xmm3= _mm_mul_ps( A._xmm3, B._xmm3 );

		return *this;
	}
	inline SSE_Float16& mult(const SSE_Float16& A, const SIMD_constant& sConst )
	{
		_xmm0= _mm_mul_ps( A._xmm0, sConst._xmm );
		_xmm1= _mm_mul_ps( A._xmm1, sConst._xmm );
		_xmm2= _mm_mul_ps( A._xmm2, sConst._xmm );
		_xmm3= _mm_mul_ps( A._xmm3, sConst._xmm );
		return *this;
	}
	inline SSE_Float16& mult(const SSE_Float16& A, const float* B)
	{
		assert( ((unsigned long)B) % 16 == 0 );
		_xmm0= _mm_mul_ps( A._xmm0, _mm_load_ps(B+0*8) );
		_xmm1= _mm_mul_ps( A._xmm1, _mm_load_ps(B+1*8) );
		_xmm2= _mm_mul_ps( A._xmm2, _mm_load_ps(B+2*8) );
		_xmm3= _mm_mul_ps( A._xmm3, _mm_load_ps(B+3*8) );

		return *this;
	}

	inline SSE_Float16& add(const SSE_Float16& A, const SSE_Float16& B)
	{
		_xmm0= _mm_add_ps( A._xmm0, B._xmm0 );
		_xmm1= _mm_add_ps( A._xmm1, B._xmm1 );
		_xmm2= _mm_add_ps( A._xmm2, B._xmm2 );
		_xmm3= _mm_add_ps( A._xmm3, B._xmm3 );

		return *this;
	}
	inline SSE_Float16& add(const SSE_Float16& A, const SIMD_constant& sConst )
	{
		_xmm0= _mm_add_ps( A._xmm0, sConst._xmm );
		_xmm1= _mm_add_ps( A._xmm1, sConst._xmm );
		_xmm2= _mm_add_ps( A._xmm2, sConst._xmm );
		_xmm3= _mm_add_ps( A._xmm3, sConst._xmm );

		return *this;
	}
	inline SSE_Float16& add(const SSE_Float16& A, const float* B)
	{
		assert( ((unsigned long)B) % 16 == 0 );
		_xmm0= _mm_add_ps( A._xmm0, _mm_load_ps(B+0*8) );
		_xmm1= _mm_add_ps( A._xmm1, _mm_load_ps(B+1*8) );
		_xmm2= _mm_add_ps( A._xmm2, _mm_load_ps(B+2*8) );
		_xmm3= _mm_add_ps( A._xmm3, _mm_load_ps(B+3*8) );

		return *this;
	}


	inline SSE_Float16& sub(const SSE_Float16& A, const SSE_Float16& B)
	{
		_xmm0= _mm_sub_ps( A._xmm0, B._xmm0 );
		_xmm1= _mm_sub_ps( A._xmm1, B._xmm1 );
		_xmm2= _mm_sub_ps( A._xmm2, B._xmm2 );
		_xmm3= _mm_sub_ps( A._xmm3, B._xmm3 );

		return *this;
	}
	inline SSE_Float16& sub(const SSE_Float16& A, const SIMD_constant& sConst )
	{
		_xmm0= _mm_sub_ps( A._xmm0, sConst._xmm );
		_xmm1= _mm_sub_ps( A._xmm1, sConst._xmm );
		_xmm2= _mm_sub_ps( A._xmm2, sConst._xmm );
		_xmm3= _mm_sub_ps( A._xmm3, sConst._xmm );

		return *this;
	}
	inline SSE_Float16& sub(const SSE_Float16& A, const float* B)
	{
		assert( ((unsigned long)B) % 16 == 0 );
		_xmm0= _mm_sub_ps( A._xmm0, _mm_load_ps(B+0*8) );
		_xmm1= _mm_sub_ps( A._xmm1, _mm_load_ps(B+1*8) );
		_xmm2= _mm_sub_ps( A._xmm2, _mm_load_ps(B+2*8) );
		_xmm3= _mm_sub_ps( A._xmm3, _mm_load_ps(B+3*8) );

		return *this;
	}


	inline SSE_Float16& fMultAdd( const SSE_Float16& mult1, const SSE_Float16& mult2 )
	{
		_xmm0= _mm_add_ps( _xmm0, _mm_mul_ps( mult1._xmm0, mult2._xmm0 ) );
		_xmm1= _mm_add_ps( _xmm1, _mm_mul_ps( mult1._xmm1, mult2._xmm1 ) );
		_xmm2= _mm_add_ps( _xmm2, _mm_mul_ps( mult1._xmm2, mult2._xmm2 ) );
		_xmm3= _mm_add_ps( _xmm3, _mm_mul_ps( mult1._xmm3, mult2._xmm3 ) );

		return *this;
	}
	inline SSE_Float16& fMultAdd( const SSE_Float16& mult1, const SIMD_constant& mult2Const  )
	{
		_xmm0= _mm_add_ps( _xmm0, _mm_mul_ps( mult1._xmm0, mult2Const._xmm ) );
		_xmm1= _mm_add_ps( _xmm1, _mm_mul_ps( mult1._xmm1, mult2Const._xmm ) );
		_xmm2= _mm_add_ps( _xmm2, _mm_mul_ps( mult1._xmm2, mult2Const._xmm ) );
		_xmm3= _mm_add_ps( _xmm3, _mm_mul_ps( mult1._xmm3, mult2Const._xmm ) );

		return *this;
	}
};



#ifdef  __AVX__

class AVX_Float16   {
protected:
	__m256 _ymm0;
	__m256 _ymm1;


public:
	int length() {return 16;}

	static const  char*  description() {return "Generic AVX, No FMA, 16 floats";}

	AVX_Float16( float* array)  //initialize
	{
		assert( ((unsigned long)array) % 32 == 0 );
		_ymm0= _mm256_load_ps( array );
		_ymm1= _mm256_load_ps( array+ 8 );
	}
	AVX_Float16( const AVX_Float16& B)  // copy constructor
	{
		_ymm0=  B._ymm0 ;
		_ymm1=  B._ymm1 ;
	}
	AVX_Float16()  // empty constructor
	{
	}
	AVX_Float16( float constVal )  // copy constructor
	{
		_ymm0= _mm256_set1_ps( constVal );
		_ymm1= _ymm0;
	}

	inline static void setConstant( SIMD_constant& sConst, const float& constVal ) {
		sConst._ymm= _mm256_set1_ps( constVal );
	}

	inline void stream( float* array)  // write bypassing cache
	{
		assert( ((unsigned long)array) % 32 == 0 );
		_mm256_stream_ps( array,     _ymm0 );
		_mm256_stream_ps( array+ 8,   _ymm1 );
	}
	inline void store( float* array)  // write including cache
	{
		assert( ((unsigned long)array) % 32 == 0 );
		_mm256_store_ps( array,     _ymm0 );
		_mm256_store_ps( array+ 8,   _ymm1 );
	}
	inline void load( float* array)  //
	{
		assert( ((unsigned long)array) % 32 == 0 );
		_ymm0= _mm256_load_ps( array      );
		_ymm1= _mm256_load_ps( array+ 8    );
	}
	inline void loadu( float* array)  //
	{
		_ymm0= _mm256_loadu_ps( array      );
		_ymm1= _mm256_loadu_ps( array+ 8    );
	}


	//  The _mm256_mul_ps() can output to a different register
	inline AVX_Float16& mult(const AVX_Float16& A, const AVX_Float16& B)
	{
		_ymm0= _mm256_mul_ps( A._ymm0, B._ymm0 );
		_ymm1= _mm256_mul_ps( A._ymm1, B._ymm1 );

		return *this;
	}
	inline AVX_Float16& mult(const AVX_Float16& A, const SIMD_constant& sConst )
	{
		_ymm0= _mm256_mul_ps( A._ymm0, sConst._ymm );
		_ymm1= _mm256_mul_ps( A._ymm1, sConst._ymm );

		return *this;
	}
	inline AVX_Float16& mult(const AVX_Float16& A, const float* B)
	{
		assert( ((unsigned long)B) % 32 == 0 );
		_ymm0= _mm256_mul_ps( A._ymm0, _mm256_load_ps(B+0*8) );
		_ymm1= _mm256_mul_ps( A._ymm1, _mm256_load_ps(B+1*8) );

		return *this;
	}

	inline AVX_Float16& add(const AVX_Float16& A, const AVX_Float16& B)
	{
		_ymm0= _mm256_add_ps( A._ymm0, B._ymm0 );
		_ymm1= _mm256_add_ps( A._ymm1, B._ymm1 );

		return *this;
	}
	inline AVX_Float16& add(const AVX_Float16& A, const SIMD_constant& sConst )
	{
		_ymm0= _mm256_add_ps( A._ymm0, sConst._ymm );
		_ymm1= _mm256_add_ps( A._ymm1, sConst._ymm );

		return *this;
	}
	inline AVX_Float16& add(const AVX_Float16& A, const float* B)
	{
		assert( ((unsigned long)B) % 32 == 0 );
		_ymm0= _mm256_add_ps( A._ymm0, _mm256_load_ps(B+0*8) );
		_ymm1= _mm256_add_ps( A._ymm1, _mm256_load_ps(B+1*8) );

		return *this;
	}


	inline AVX_Float16& sub(const AVX_Float16& A, const AVX_Float16& B)
	{
		_ymm0= _mm256_sub_ps( A._ymm0, B._ymm0 );
		_ymm1= _mm256_sub_ps( A._ymm1, B._ymm1 );

		return *this;
	}
	inline AVX_Float16& sub(const AVX_Float16& A, const SIMD_constant& sConst )
	{
		_ymm0= _mm256_sub_ps( A._ymm0, sConst._ymm );
		_ymm1= _mm256_sub_ps( A._ymm1, sConst._ymm );

		return *this;
	}
	inline AVX_Float16& sub(const AVX_Float16& A, const float* B)
	{
		assert( ((unsigned long)B) % 32 == 0 );
		_ymm0= _mm256_sub_ps( A._ymm0, _mm256_load_ps(B+0*8) );
		_ymm1= _mm256_sub_ps( A._ymm1, _mm256_load_ps(B+1*8) );

		return *this;
	}

	// we can call the other functions for AVX
	inline AVX_Float16& operator*=( const AVX_Float16& B)
      {
		return this->mult( *this , B);
      }

	inline AVX_Float16& operator*=( const float* B)
      {
		assert( ((unsigned long)B) % 32 == 0 );
		return this->mult( *this , B);
      }
	inline AVX_Float16& operator*=( const SIMD_constant& sConst  )
      {
		return this->mult( *this , sConst);
      }

	inline AVX_Float16& operator+=( const AVX_Float16& B)
      {
		_ymm0= _mm256_add_ps( _ymm0, B._ymm0 );
		_ymm1= _mm256_add_ps( _ymm1, B._ymm1 );

		return *this;
      }
	inline AVX_Float16& operator+=( const float* B)
      {
		assert( ((unsigned long)B) % 32 == 0 );
		_ymm0= _mm256_add_ps( _ymm0, _mm256_load_ps(B+0*8) );
		_ymm1= _mm256_add_ps( _ymm1, _mm256_load_ps(B+1*8) );

		return *this;
      }
	inline AVX_Float16& operator+=( const SIMD_constant& sConst  )
    {
		_ymm0= _mm256_add_ps( _ymm0, sConst._ymm );
		_ymm1= _mm256_add_ps( _ymm1, sConst._ymm );

		return *this;
    }

	inline AVX_Float16& operator-=( const AVX_Float16& B)
      												{
		_ymm0= _mm256_sub_ps( _ymm0, B._ymm0 );
		_ymm1= _mm256_sub_ps( _ymm1, B._ymm1 );

		return *this;
      												}
	inline AVX_Float16& operator-=( const float* B)
      												{
		assert( ((unsigned long)B) % 32 == 0 );
		_ymm0= _mm256_sub_ps( _ymm0, _mm256_load_ps(B+0*8) );
		_ymm1= _mm256_sub_ps( _ymm1, _mm256_load_ps(B+1*8) );

		return *this;
      												}
	inline AVX_Float16& operator-=( const SIMD_constant& sConst  )
      												{
		_ymm0= _mm256_sub_ps( _ymm0, sConst._ymm );
		_ymm1= _mm256_sub_ps( _ymm1, sConst._ymm );

		return *this;
      												}


	inline AVX_Float16& fMultAdd( const AVX_Float16& mult1, const AVX_Float16& mult2 )
	{
		_ymm0= _mm256_add_ps( _ymm0, _mm256_mul_ps( mult1._ymm0, mult2._ymm0 ) );
		_ymm1= _mm256_add_ps( _ymm1, _mm256_mul_ps( mult1._ymm1, mult2._ymm1 ) );
		return * this;
	}
	inline AVX_Float16& fMultAdd( const AVX_Float16& mult1, const SIMD_constant& sConst )
	{
		_ymm0= _mm256_add_ps( _ymm0, _mm256_mul_ps( mult1._ymm0, sConst._ymm ) );
		_ymm1= _mm256_add_ps( _ymm1, _mm256_mul_ps( mult1._ymm1, sConst._ymm ) );
		return * this;
	}

};
#endif // __AVX__


#ifdef __FMA4__
class AVX_FMA4_Float16 : public AVX_Float16  {
public:
	static const char*  description() {return "AVX, AMD FMA, 16 floats";}

	AVX_FMA4_Float16(  float* array) : AVX_Float16(array) {};
	AVX_FMA4_Float16(  float c) : AVX_Float16(c) {};
	AVX_FMA4_Float16( const AVX_FMA4_Float16& B)  // copy constructor
	{
		_ymm0=  B._ymm0 ;
		_ymm1=  B._ymm1 ;
	}
	AVX_FMA4_Float16()  // empty constructor
	{
	}

	inline AVX_FMA4_Float16& fMultAdd( const AVX_FMA4_Float16& mult1, const AVX_FMA4_Float16& mult2 )
	{
		_ymm0= _mm256_macc_ps( mult1._ymm0, mult2._ymm0, _ymm0 );
		_ymm1= _mm256_macc_ps( mult1._ymm1, mult2._ymm1, _ymm1 );
		return * this;
	}
	inline AVX_FMA4_Float16& fMultAdd( const AVX_FMA4_Float16& mult1, const SIMD_constant& sConst )
	{
		_ymm0= _mm256_macc_ps( mult1._ymm0, sConst._ymm, _ymm0 );
		_ymm1= _mm256_macc_ps( mult1._ymm1, sConst._ymm, _ymm1 );
		return * this;
	}
};
#endif

#ifdef __FMA__
class AVX_FMA3_Float16 : public AVX_Float16  {
public:
	static const  char*  description() {return "AVX, Intel FMA, 16 floats";}

	AVX_FMA3_Float16(  float* array) : AVX_Float16(array) {};
	AVX_FMA3_Float16(  float c) : AVX_Float16(c) {};
	AVX_FMA3_Float16( const AVX_FMA3_Float16& B)  // copy constructor
	{
		_ymm0=  B._ymm0 ;
		_ymm1=  B._ymm1 ;
	}
	AVX_FMA3_Float16()  // empty constructor
	{
	}


	inline AVX_FMA3_Float16& fMultAdd( const AVX_FMA3_Float16& mult1, const AVX_FMA3_Float16& mult2 )
	{
		_ymm0= _mm256_fmadd_ps( mult1._ymm0, mult2._ymm0, _ymm0 );
		_ymm1= _mm256_fmadd_ps( mult1._ymm1, mult2._ymm1, _ymm1 );
		return * this;
	}
	inline AVX_FMA3_Float16& fMultAdd(  const AVX_FMA3_Float16& mult1, const SIMD_constant& sConst )
	{
		_ymm0= _mm256_fmadd_ps( mult1._ymm0, sConst._ymm, _ymm0 );
		_ymm1= _mm256_fmadd_ps( mult1._ymm1, sConst._ymm, _ymm1 );
		return * this;
	}
};
#endif




static void  cpuid( int result[4], int idnum) {
#ifdef __INTEL_COMPILER
	__cpuid(result, idnum );
#else  // __INTEL_COMPILER
	int  i0, i1, i2, i3;
	__asm("cpuid":  "=a"(i0), "=b"(i1), "=c"(i2), "=d"(i3) : "a"(idnum), "c"(0) : );
	result[0]= i0;
	result[1]= i1;
	result[2]= i2;
	result[3]= i3;
#endif // __INTEL_COMPILER
}

// this is needed because the cpu may support AVX, but the OS may not.
static long  xgetbv(int iarg) {
#ifndef __AVX__
	return 0;
#else
#ifdef __INTEL_COMPILER
	int i0;
	i0= _xgetbv(iarg);
	return i0;
#else  // __INTEL_COMPILER
	int i0, i1;
	__asm("xgetbv":  "=a"(i0), "=d"(i1) : "c"(iarg) : );
	//long l2= ((long)i1) << 32;
	//return ( i0 | l2 );
	return  i0 ;
#endif // __INTEL_COMPILER
#endif
}

static bool AVX_works(void) {
	static int ireturn= -1;
	if ( ireturn >= 0 )
		return ireturn;

	int cpuflags[4];
	cpuid(cpuflags,1);

	ireturn= 0;
	if ( (cpuflags[2] & (0x1<<27) != 0) &&
			(cpuflags[2] & (0x1<<28) !=0) &&
			(xgetbv(0) & 0x6 == 0x6) )
		ireturn= 1;

	return ireturn;
}

static bool SSE_works(void) {
	static int ireturn= -1;
	if ( ireturn >= 0 )
		return ireturn;

	int cpuflags[4];
	cpuid(cpuflags,1);

	ireturn= 0;
	if ( (cpuflags[3] & (0x1<<25) != 0)  )
		ireturn= 1;

	return ireturn;
}

static bool FMA4_works(void) {
	static int ireturn= -1;
	if ( ireturn >= 0 )
		return ireturn;

	ireturn= 0;
	if ( AVX_works() ) {
		int cpuflags[4];
		cpuid( cpuflags, 0x80000001);
		if ( cpuflags[2] & (0x1<<16) != 0 )
			ireturn= 1;
	}
	return ireturn;
}

static bool FMA3_works(void) {
	static int ireturn= -1;
	if ( ireturn >= 0 )
		return ireturn;

	ireturn= 0;
	if ( AVX_works() ) {
		int cpuflags[4];
		cpuid( cpuflags, 1);
		if ( cpuflags[2] & (0x1<<12) != 0 )
			ireturn= 1;
	}
	return ireturn;
}


//#endif /* FLEXSIMD_H_ */

